#!/usr/bin/env python3

import freezerbox
import parametrize_from_file
import inform

from stepwise import Q, pl, table, format_text
from freezerbox import Database, parse_fields, parse_fields_list, cd
from freezerbox.stepwise.dilute import *
from schema_helpers import *
from mock_model import *
from pytest import approx
from math import inf

given_expected_error = Schema({
    'given': eval,
     **error_or({
         'expected': eval_with(Stock=Stock, Q=Q, approx=approx),
     }),
})
app_expected_error=Schema({
    'app': exec_with('app', Dilute=Dilute, Stock=Stock, Q=Q),
    **error_or({
        'expected': eval_with(
            Q=Q,
            approx=approx,
            approx_Q=approx_Q,
            pl=pl,
            table=table,
        ),
    }),
})
db_app_expected_error=Schema({
    'db': eval_db,
    **app_expected_error.schema,
})
app_expected_error_stderr=Schema({
    Optional('stderr', default=''): str,
    **app_expected_error.schema,
})

@parametrize_from_file(schema=given_expected_error)
def test_parse_stock(given, expected, error):
    with error, cd(TEST_DIR):
        assert list(parse_stock(given)) == expected

@parametrize_from_file(schema=given_expected_error)
def test_parse_conc(given, expected, error):
    with error:
        assert parse_conc(given) == expected

@parametrize_from_file(schema=given_expected_error)
def test_parse_mw(given, expected, error):
    with error:
        assert parse_mw(given) == expected

@parametrize_from_file(schema=given_expected_error)
def test_parse_volume(given, expected, error):
    with error:
        assert parse_volume(given) == approx(expected)

@parametrize_from_file(schema=db_app_expected_error)
def test_dilute_concs(db, app, expected, error):
    with error:
        app.db = db
        assert app.concs.to_dict('records') == expected

@parametrize_from_file(schema=app_expected_error_stderr)
def test_dilute_dilutions(app, expected, error, stderr, capsys):
    # The default informer has already stored references to `sys.stdout` and 
    # `sys.stderr`, so it won't be affected by `capsys`.  In order to actually 
    # capture these streams, we need to make a new informer within the test.

    # Copied from `stepwise/__init__.py`:
    informer = inform.Inform(stream_policy='header')

    # Don't accidentally use the database present on the testing system.
    app.db = freezerbox.Database()

    with error:
        dilutions = app.dilutions[['stock_uL', 'diluent_uL', 'final_conc']]
        assert dilutions.to_dict('records') == expected

        cap = capsys.readouterr()
        assert stderr in cap.err

    informer.disconnect()

@parametrize_from_file(schema=db_app_expected_error)
def test_dilute_protocol(db, app, expected, error):
    expected['footnotes'] = {
            int(k): v
            for k, v in expected['footnotes'].items()
    }
    app.db = db

    with error:
        p = app.protocol

        print("EXPECTED STEPS")
        print('\n\n'.join(format_text(x, inf) for x in expected['steps']))
        print()
        print("EXPECTED FOOTNOTES")
        print('\n\n'.join(format_text(x, inf) for x in expected['footnotes'].values()))
        print()
        print("ACTUAL")
        print(p.format_text())

        assert p.steps == expected['steps']
        assert p.footnotes == expected['footnotes']

def test_make(mock_plugins):
    db = Database()
    db['x1'] = MockMolecule(
            synthesis=parse_fields('a'),
            cleanups=parse_fields_list('dilute conc=10nM'),
    )
    assert db['x1'].conc_nM == 10




#!/usr/bin/env python3

import pytest
from freezerbox import load_db, parse_tag, Fields
from freezerbox.model import *
from pathlib import Path
from schema_helpers import *
from mock_model import *
from datetime import datetime

MOCK_DB = Path(__file__).parent / 'mock_excel_db'

@pytest.fixture(autouse=True)
def db(mock_plugins):
    MOCK_CONFIG = {
            'use': 'mock',
            'database': {
                'mock': {
                    'type': 'excel',
                    'dir': str(MOCK_DB),
                }
            }
    }
    return load_db(config=MOCK_CONFIG)


@parametrize_from_file(
        schema=Schema({
            'path': str,
            'tag': str,
            'expected': eval,
        }),
)
def test_path_from_tag(tmp_path, path, tag, expected):
    from freezerbox.loaders.excel import _path_from_tag

    path = tmp_path / path
    path.touch()

    tag = parse_tag(tag)
    hit = _path_from_tag(tmp_path, tag)

    if expected:
        assert hit == path
    else:
        assert hit == None


def test_f2(db, tag='f2'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].alt_names == ['name 1', 'name 2']
    assert db[tag].date == datetime(2021, 4, 19)
    assert db[tag].desc == "description"
    assert db[tag].molecule == 'DNA'
    assert db[tag].is_double_stranded == True
    assert db[tag].is_circular == False
    assert db[tag].ready == True

def test_f3(db, tag='f3'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "sequence from db"
    assert db[tag].seq == 'GAATTC'

def test_f4(db, tag='f4'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "sequence from file"
    assert db[tag].seq == 'AAGCTT'

def test_f5(db, tag='f5'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "sequence from synthesis"
    assert db[tag].synthesis_args == Fields(['mock'], {'seq': 'GGTCTC'})
    assert db[tag].seq == 'GGTCTC'

def test_f6(db, tag='f6'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "mw from db"
    assert db[tag].mw == pytest.approx(5000, abs=0.1)

def test_f7(db, tag='f7'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "mw from sequence"
    # http://molbiotools.com/dnacalculator.html
    assert db[tag].mw == pytest.approx(3582.45, abs=0.1)

def test_f8(db, tag='f8'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "conc from db"
    assert db[tag].conc_nM == 60

def test_f9(db, tag='f9'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "conc from cleanup"
    assert db[tag].synthesis_args == Fields(['mock'], {})
    assert db[tag].cleanup_args == [Fields(['mock'], {'conc': '50nM'})]
    assert db[tag].conc_nM == 50

def test_f10(db, tag='f10'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "molecule from db"
    assert db[tag].molecule == 'RNA'
    assert db[tag].is_single_stranded == True

def test_f11(db, tag='f11'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "molecule from synthesis"
    assert db[tag].synthesis_args == Fields(['mock'], {'molecule': 'ssRNA'})
    assert db[tag].molecule == 'RNA'
    assert db[tag].is_single_stranded == True

def test_f12(db, tag='f12'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "length from db"
    assert db[tag].length == 10

def test_f13(db, tag='f13'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "length from sequence"
    assert db[tag].seq == "GAATTC"
    assert db[tag].length == 6

def test_f14(db, tag='f14'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "circular from db"
    assert db[tag].is_circular == True

def test_f15(db, tag='f15'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "circular from synthesis"
    assert db[tag].synthesis_args == Fields(['mock'], {'circular': 'y'})
    assert db[tag].is_circular == True

def test_f16(db, tag='f16'):
    assert isinstance(db[tag], NucleicAcid)
    assert db[tag].desc == "ready from db"
    assert db[tag].ready == False


def test_o2(db, tag='o2'):
    assert isinstance(db[tag], Oligo)
    assert db[tag].name == 'o2_tm61'
    assert db[tag].desc == 'tm from name'
    assert db[tag].tm == pytest.approx(61, abs=0.1)

def test_o3(db, tag='o3'):
    assert isinstance(db[tag], Oligo)
    assert db[tag].desc == 'tm from sequence'
    assert db[tag].seq == 'TCTCGCGGTATCATTG'
    assert db[tag].tm == pytest.approx(48, abs=0.1)

def test_o4(db, tag='o4'):
    assert isinstance(db[tag], Oligo)
    assert db[tag].desc == 'ignore scale purification'


def test_p2(db, tag='p2'):
    assert isinstance(db[tag], Plasmid)
    assert db[tag].is_double_stranded
    assert db[tag].is_circular


def test_r2(db, tag='r2'):
    assert isinstance(db[tag], Protein)


def test_b2(db, tag='b2'):
    assert isinstance(db[tag], Buffer)



#!/usr/bin/env python3

from itertools import groupby, count
from more_itertools import unzip
from statistics import mean
from operator import attrgetter
from ..utils import join_lists

# group_by_...(items, key):
#   Functions that yield (value, subitems) tuples.  Each item in `subitems` 
#   will be associated with the given value.

# Items will be maker instances

def iter_combo_makers(
        factory,
        solo_makers, *,
        group_by={},
        merge_by={},
        keys={},
    ):
    combos = iter_combos(
            solo_makers,
            group_by=group_by,
            merge_by={
                'products': join_lists,
                **merge_by,
            },
            keys=keys,
    )
    for attrs, items in combos:
        combo_maker = factory()
        for k, v in attrs.items():
            setattr(combo_maker, k, v)
        yield combo_maker

def iter_combos(items, *, group_by={}, merge_by={}, keys={}):

    def get_key(attr):
        return keys.get(attr, lambda x: getattr(x, attr))

    def do_iter_combos(items, group_by, attrs):
        if not items:
            return

        if not group_by:
            items = list(items)

            for attr, merge in merge_by.items():
                key = get_key(attr)
                values = map(key, items)
                attrs[attr] = merge(values)

            yield attrs, items
            return

        x = list(group_by.items())
        head, tail = x[0], x[1:]

        attr, grouper = head
        group_by_tail = dict(tail)
        key = get_key(attr)

        for group_value, group_items in grouper(items, key=key):
            yield from do_iter_combos(
                    group_items,
                    group_by_tail,
                    {**attrs, attr: group_value},
            )

    yield from do_iter_combos(items, group_by, {})

class group_by_cluster:

    def __init__(self, *, max_diff, aggregate=mean):
        self.max_diff = max_diff
        self.aggregate = aggregate

    def __call__(self, items, key=lambda x: x):
        import numpy as np
        from sklearn.cluster import AgglomerativeClustering

        if len(items) == 0:
            return
        if len(items) == 1:
            value = self.aggregate(map(key, items))
            yield value, items
            return

        clustering = AgglomerativeClustering(
                linkage='complete',
                n_clusters=None,
                distance_threshold=self.max_diff,
        )

        # Specify `dtype=object` to prevent numpy from casting the keys, which 
        # in some cases can lead to confusing results (e.g. `np.int64` doesn't 
        # undergo true division with `statistics.mean()`).
        keys = map(key, items)
        keys = np.array(list(keys), dtype=object).reshape(-1, 1)

        labels = clustering.fit_predict(keys)

        by_label = lambda x: x[0]
        item_infos = sorted(
                zip(labels, items, keys.flat, count()),
                key=by_label,
        )

        # Note that we take care to yield the clustered items in roughly the 
        # same order as they were given to us.

        clusters = []

        for label, group in groupby(item_infos, key=by_label):
            _, group_items, group_keys, group_indices = unzip(group)
            clusters.append((
                    mean(group_indices),
                    self.aggregate(group_keys),
                    list(group_items),
            ))

        for order, value, items in sorted(clusters):
            yield value, items

def group_by_identity(items, key=lambda x: x):
    # The advantage of this algorithm is that it only requires the items to 
    # support the equality operator.  More efficient algorithms would require 
    # additional functionality, and therefore would not support all inputs:
    # 
    # - itertools.groupby(): inputs must be sorted, therefore items must 
    #   support comparison operators.
    # - dict of lists: items must be hashable.

    groups = []

    for item in items:
        item_key = key(item)

        for group in groups:
            if item_key == group[0]:
                group[1].append(item)
                break
        else:
            groups.append((item_key, [item]))

    yield from groups




#!/usr/bin/env python3
# encoding: utf-8

from setuptools import setup

with open('README.rst') as file:
    readme = file.read()

setup(
    name='mock_project',
    version='0.0.0',
    author='Kale Kundert',
    long_description=readme,
    packages=[
        'mock_project',
    ],
    install_requires=[
    ],
    entry_points={
        'console_scripts': [
        ],
    },
    include_package_data=True,
)



#!/usr/bin/env python3

import freezerbox
import pytest
import parametrize_from_file

from freezerbox.fields import \
        Fields, word, key_value, parse_fields, parse_fields_list, _quote_word
from freezerbox.errors import ParseError
from parsy import ParseError as ParsyError
from schema_helpers import *

fields_dict = empty_ok({
    'by_index': empty_ok([str]),
    'by_name': empty_ok({str: str}),
})
fields_dict_list = empty_ok([fields_dict])


@parametrize_from_file
def test_word(given, expected):
    assert word.parse(given) == expected

@parametrize_from_file
def test_word_err(given):
    with pytest.raises(ParsyError):
        word.parse(given)

@parametrize_from_file
def test_key_value(given, expected):
    assert key_value.parse(given) == (expected['key'], expected['value'])

@parametrize_from_file
def test_key_value_err(given):
    with pytest.raises(ParsyError):
        key_value.parse(given)

@parametrize_from_file(
        schema=Schema({
            'given': str,
            'expected': fields_dict,
        }),
)
def test_fields(given, expected):
    parsed = parse_fields(given)
    assert parsed.by_index == expected['by_index']
    assert parsed.by_name == expected['by_name']

@parametrize_from_file
def test_fields_err(given, messages):
    with pytest.raises(ParseError) as err:
        parse_fields(given)

    for message in messages:
        assert message in str(err.value)

@parametrize_from_file(
        schema=Schema({
            'given': str,
            'expected': fields_dict_list,
        }),
)
def test_fields_list(given, expected):
    parsed = [
            {'by_index': x.by_index, 'by_name': x.by_name}
            for x in parse_fields_list(given)
    ]
    assert parsed == expected

@parametrize_from_file
def test_fields_list_err(given, messages):
    with pytest.raises(ParseError) as err:
        parse_fields_list(given)

    for message in messages:
        assert message in str(err.value)

@parametrize_from_file
def test_fields_repr(expr):
    fields = eval_freezerbox(expr)
    assert repr(fields) == expr

@parametrize_from_file
def test_fields_str(given, expected):
    fields = eval_freezerbox(given)
    assert str(fields) == expected


def test_fields_eq():
    f1 = Fields(['a'], {'b': 2})
    f2 = Fields(['a'], {'b': 2})
    g1 = Fields(['x'], {'b': 2})
    g2 = Fields(['a'], {'x': 2})
    g3 = Fields(['a'], {'b': 'x'})

    assert f1 == f2
    assert f1 != g1
    assert f1 != g2
    assert f1 != g3

def test_fields_getitem_get():
    fields = Fields(['a', 'b'], {'c': 3, 'd': 4})

    assert fields[0] == fields.get(0) == 'a'
    assert fields[1] == fields.get(1) == 'b'

    assert fields.get(2) == None
    with pytest.raises(KeyError):
        fields[2]

    assert fields['c'] == fields.get('c') == 3
    assert fields['d'] == fields.get('d') == 4

    assert fields.get('e') == None
    with pytest.raises(KeyError):
        fields['e']

def test_fields_contains():
    f1 = Fields([], {})

    assert 0 not in f1
    assert 'a' not in f1

    f2 = Fields(['a'], {'b': 2})
    assert 0 in f2
    assert 1 not in f2
    assert 2 not in f2
    assert 'a' not in f2
    assert 'b' in f2

    f3 = Fields(['a', 'b'], {'c': 3, 'd': 4})

    assert 0 in f3
    assert 1 in f3
    assert 3 not in f3
    assert 4 not in f3
    assert 'a' not in f3
    assert 'b' not in f3
    assert 'c' in f3
    assert 'd' in f3

@parametrize_from_file
def test_quote_word(given, expected):
    assert _quote_word(given) == expected



#!/usr/bin/env python3

import pandas as pd
import autosnapgene as snap
from pathlib import Path
from voluptuous import Required, All, Any, Coerce
from warnings import catch_warnings, filterwarnings
from stepwise import Quantity
from ..model import Database, Tag
from ..model import NucleicAcid, Protein, Buffer, Plasmid, Oligo
from ..fields import parse_fields, parse_fields_list
from ..utils import *

schema = {
        'type': 'excel',
        Required('dir'): All(str, Coerce(Path)),
        'columns': {str: Any(
            'seq', 'molecule', 'synthesis', 'cleanups', 'ready', 'name',
            'alt_names', 'date', 'desc', 'length', 'conc', 'mw', 'circular',
        )},
}
default_config = {
        'columns': {
            'Sequence': 'seq',
            'Molecule': 'molecule',
            'Synthesis': 'synthesis',
            'Cleanups': 'cleanups',
            'Ready': 'ready',
            'Name': 'name',
            'Cross-refs': 'alt_names',
            'Date': 'date',
            'Description': 'desc',
            'Length': 'length',
            'Conc': 'conc',
            'MW': 'mw',
            'Circular': 'circular',
        }
}

def load(config):
    root = config['dir']

    db_xlsx = {
            NucleicAcid: root / 'fragments.xlsx',
            Plasmid:     root / 'plasmids.xlsx',
            Oligo:       root / 'oligos.xlsx',
            Protein:     root / 'proteins.xlsx',
            Buffer:      root / 'buffers.xlsx',
    }
    seq_dirs = {
            NucleicAcid: root / 'fragments',
            Plasmid:     root / 'plasmids',
            Oligo:       root / 'oligos',
            Protein:     root / 'proteins',
    }

    for dir in seq_dirs.values():
        dir.mkdir(exist_ok=True, parents=True)

    db = Database(str(root))

    for cls, path in db_xlsx.items():
        if not path.exists():
            continue

        with catch_warnings():
            filterwarnings(
                    'ignore',
                    category=UserWarning,
                    message="Workbook contains no default style, apply openpyxl's default",
            )
            filterwarnings(
                    'ignore',
                    category=DeprecationWarning,
                    message="`np.float` is a deprecated alias for the builtin `float`.",
            )
            df = pd.read_excel(path)

        df = df.set_index(df.index + 2)
        df = df.rename(columns=config['columns'])
        df = df.astype(object).where(pd.notnull(df), None)

        for i, row in df.iterrows():
            tag = Tag(cls.tag_prefix, i)
            kwargs = {k: v for k, v in row.items() if v is not None}

            if not kwargs.get('seq') and cls is not Buffer:
                kwargs['seq'] = _defer(_seq_from_tag, seq_dirs[cls], tag)
            if x := kwargs.get('alt_names'):
                kwargs['alt_names'] = [y.strip() for y in x.split(',')]
            if x := kwargs.get('conc'):
                kwargs['conc'] = _defer(Quantity.from_string, x)
            if x := kwargs.get('synthesis'):
                kwargs['synthesis'] = _defer(parse_fields, x)
            if x := kwargs.get('cleanups'):
                kwargs['cleanups'] = _defer(parse_fields_list, x)
            if x := kwargs.get('circular'):
                kwargs['circular'] = _defer(parse_bool, x)
            if x := kwargs.get('ready'):
                kwargs['ready'] = _defer(parse_bool, x)

            db[tag] = cls(**kwargs)

    return db

def _defer(f, *args, **kwargs):
    """
    Create a closure that calls the given function with the given arguments.

    The point of this function is to avoid the surprising behavior that can 
    occur if you define a closure in a scope where variables are changing (e.g. 
    in a for-loop).  The confusing thing is that closures have access to the 
    scope they were defined in, but only as it exists when they are ultimately 
    called.  So if the scope changes between when the closure is defined and 
    when it's called, the closure will use the final value of any variables.

    This function serves to create a static local scope containing the 
    variables needed by the closure, which avoids the problem.

    More information: 
    https://stackoverflow.com/questions/10452770/python-lambdas-binding-to-local-values
    """
    class defer:

        def __repr__(self):
            return '<deferred>'

        def __call__(self):
            return f(*args, **kwargs)

    return defer()

def _seq_from_tag(dir, tag):
    if path := _path_from_tag(dir, tag):
        return snap.parse(path).dna_sequence

def _path_from_tag(dir, tag):
    # Figure out the path to the actual file, allowing for several different 
    # naming conventions.
    for path in dir.iterdir():
        if re.fullmatch(f'({tag.type})?0*{tag.id}([_-].*)?.dna', path.name):
            return path




#!/usr/bin/env python3

"""
A simple format for extracting fields from a string.
"""

from .errors import ParseError
from parsy import (
        ParseError as ParsyError,
        generate, regex, string, alt, test_char,
)
from contextlib import contextmanager

QUOTES = {'"': '"', "'": "'", '“': '”'}

class Fields:

    def __init__(self, indexed, named):
        self.by_index = indexed
        self.by_name = named

    def __repr__(self):
        return f'Fields({self.by_index!r}, {self.by_name!r})'

    def __str__(self):
        index_strs = map(_quote_word, self.by_index)
        name_strs = (f'{k}={_quote_word(v)}' for k, v in self.by_name.items())
        return ' '.join((*index_strs, *name_strs))

    def __eq__(self, other):
        return self.by_index == other.by_index and \
               self.by_name == other.by_name

    def __contains__(self, key):
        if isinstance(key, str):
            return key in self.by_name

        if isinstance(key, int):
            return key < len(self.by_index)

        return False

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.by_name[key]

        if isinstance(key, int):
            try:
                return self.by_index[key]
            except IndexError:
                raise KeyError(key)

        raise KeyError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError as err:
            return default

def parse_fields_list(fields_str):
    with _reformat_errors():
        return fields_list.parse(fields_str)
        
def parse_fields(fields_str):
    with _reformat_errors():
        return fields.parse(fields_str)

@generate
def fields_list():
    sep = regex(r'\s*;\s+')
    parsed = yield fields.sep_by(sep)
    return parsed

@generate
def fields():
    by_index = []
    by_name = {}
    space = regex(r'\s+')

    @generate
    def field():
        # Always try matching named fields before indexed fields.  This is 
        # necessary the parser for an indexed field will match the first half 
        # of a named field but stop at the '='.  The next parser will then fail 
        # when it tries to continue from there.
        if by_name:
            yield named_field
        else:
            yield named_field | indexed_field

    @generate
    def named_field():
        k, v = yield key_value
        by_name[k] = v

    @generate
    def indexed_field():
        v = yield word
        by_index.append(v)
    
    yield field.sep_by(space, min=1)
    return Fields(by_index, by_name)

@generate
def key_value():
    key = yield word
    yield string('=')
    value = yield word
    return key, value

@generate
def word():
    open_quote_char = yield alt(*map(string, QUOTES)).desc("quote").optional()

    if open_quote_char is None:
        return unquoted_word

    else:
        close_quote_char = QUOTES[open_quote_char]
        escape = regex(fr'\\[\\{close_quote_char}]').map(lambda x: x[-1]).desc("escaped quote")
        value_char = escape | regex(fr'[^\\{close_quote_char}]+')
        word = yield value_char.many().concat()
        yield string(close_quote_char).desc("quote")
        return word

unquoted_word = regex(r'''[^=;\s'"]+''').desc('word')

@contextmanager
def _reformat_errors():
    try:
        yield

    except ParsyError as e1:
        e2 = ParseError(
                stream=e1.stream,
                index=e1.index,
                expected=e1.expected,
        )

        def format_stream(e):
            return f"parsing:  {e.stream!r}"

        def format_expected(e):
            expected = ' or '.join(
                    sorted(map(str, e.expected)),
            )
            return f"expected:  {e.index * ' '}^ {expected}"

        e2.brief = "can't parse fields"
        e2.info += format_stream
        e2.blame += format_expected

        raise e2 from e1

def _quote_word(x):
    x = str(x)
    try:
        unquoted_word.parse(x)
    except ParsyError:
        x = x.replace('\\', '\\\\').replace("'", r"\'")
        return f"'{x}'"
    else:
        return x




#!/usr/bin/env python3

import freezerbox, pytest
import parametrize_from_file

from pathlib import Path
from freezerbox import load_config, ReagentConfig, MakerConfig, Fields, cd
from appcli.model import Log
from more_itertools import one, first, zip_equal
from test_model import MockReagent
from re_assert import Matches
from operator import attrgetter
from mock_model import mock_plugins
from pprint import pprint
from schema_helpers import *

MOCK_CONFIG = Path(__file__).parent / 'mock_config'

class MockObj:
    pass

def test_config():
    with cd(MOCK_CONFIG):
        load_config.cache_clear()
        config = load_config()

        # Only check the values that are explicitly set by the test, because 
        # any other values could be affected by real configuration files 
        # present on the tester's system.

        assert config['use'] == 'db1'
        assert config['database']['db1']['type'] == 'type1'
        assert config['database']['db1']['option'] == 'option1'
        assert config['database']['db2']['type'] == 'type2'
        assert config['database']['db2']['option'] == 'option2'

    with cd(MOCK_CONFIG / 'subdir'):
        load_config.cache_clear()
        config = load_config()

        assert config['use'] == 'db2'
        assert config['database']['db1']['type'] == 'type1'
        assert config['database']['db1']['option'] == 'option1'
        assert config['database']['db2']['type'] == 'type2'
        assert config['database']['db2']['option'] == 'option2'

    load_config.cache_clear()

@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            'obj': exec_with(
                'obj',
                MockObj=MockObj,
            ),
            Optional('config_cls', default='class MockConfig(ReagentConfig): pass'): exec_with(
                'MockConfig',
                ReagentConfig=freezerbox.ReagentConfig,
                attrgetter=attrgetter,
            ),
            Optional('db_access', default='cache'): str,
            'key': eval_with(
                attrgetter=attrgetter,
            ),
            'expected': empty_ok([eval_python]),
            'info': [str],
    }),
)
def test_reagent_config(db, config_cls, obj, db_access, key, expected, info, monkeypatch, mock_plugins):
    db.name = '/path/to/db'
    config = config_cls(obj)
    layer = one(config.load())

    if db_access == 'cache':
        layer.db = db

    if db_access.startswith('obj'):
        attr = db_access.split('.')[1] if '.' in db_access else 'db'
        setattr(obj, attr, db)

    if db_access == 'load':
        monkeypatch.setattr(freezerbox.model, 'load_db', lambda: db)
    else:
        monkeypatch.setattr(freezerbox.model, 'load_db', lambda: NotImplemented)

    log = Log()
    values = list(layer.iter_values(key, log))

    assert values == expected

    for actual, pattern in zip_equal(log.err.info_strs, info):
        Matches(pattern).assert_matches(actual)

@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            'config_cls': exec_with(
                'MockConfig',
                ProductConfig=freezerbox.ProductConfig,
                MakerConfig=freezerbox.MakerConfig,
                PrecursorConfig=freezerbox.PrecursorConfig,
                attrgetter=attrgetter,
            ),
            'products': {str: Coerce(int)},
            Optional('products_attr', default='products'): str,
            'key': eval_with(
                attrgetter=attrgetter,
            ),
            **error_or({
                'expected': empty_ok([eval_freezerbox]),
                'info': [str],
            }),
    }),
)
def test_product_configs(db, config_cls, products, products_attr, key, expected, info, error, mock_plugins):
    db.name = 'path/to/db'
    if not products:
        products = db.keys()

    obj = MockObj()

    if products_attr:
        setattr(obj, products_attr, [
            db[k].make_intermediate(i)
            for k,i in products.items()
        ])

    config = config_cls(obj)
    layer = one(config.load())

    with error:
        log = Log()
        values = list(layer.iter_values(key, log))
        pprint(log.err.info_strs)

        assert values == expected

        for actual, pattern in zip_equal(log.err.info_strs, info):
            Matches(pattern).assert_matches(actual)




#!/usr/bin/env python3

import freezerbox
import pytest
import parametrize_from_file
import networkx as nx
import stepwise

from stepwise import Quantity, Q
from voluptuous import Schema, Invalid, Coerce, And, Or, Optional
from unittest.mock import MagicMock
from mock_model import *
from contextlib import nullcontext
from pathlib import Path

TEST_DIR = Path(__file__).parent

class do_with:

    def __init__(self, globals=None, **kw_globals):
        self.globals = globals or {}
        self.globals.update(kw_globals)

    def all(self, module):
        try:
            keys = module.__all__
        except AttributeError:
            keys = module.__dict__

        self.globals.update({
                k: module.__dict__[k]
                for k in keys
        })
        return self

class eval_with(do_with):

    def __call__(self, src):
        try:
            if isinstance(src, list):
                return [self(x) for x in src]
            elif isinstance(src, dict):
                return {k: self(v) for k, v in src.items()}
            else:
                return eval(src, self.globals)

        except Exception as err:
            raise Invalid(str(err)) from err

class exec_with(do_with):

    def __init__(self, get, globals=None, **kw_globals):
        super().__init__(globals, **kw_globals)
        self.get = get

    def __call__(self, src):
        globals = self.globals.copy()

        try:
            exec(src, globals)
        except Exception as err:
            raise Invalid(str(err)) from err

        if callable(self.get):
            return self.get(globals)
        else:
            return globals[self.get]

def eval_db(reagents):
    db = freezerbox.Database()
    schema = Schema(
            empty_ok({
                str: eval_freezerbox,
            }),
    )

    for tag, reagent in schema(reagents).items():
        db[tag] = reagent

    return db

def empty_ok(x):
    return Or(x, And('', lambda y: type(x)()))

def error_or(expected):
    schema = {}

    # Either specify an error or an expected value, not both.
    # KBK: This doesn't work for some reason.
    #schema[Or('error', *expected, only_one=True)] = object

    schema[Optional('error', default='none')] = error

    def check(schema):
        # This function basically just reimplements `Or`, but with a better 
        # error message.  The error message for `Or` is confusing because it 
        # only mentions the first option and not the second.  That's especially 
        # bad in this case, where the first option is basically an internal 
        # implementation detail.  This function changes the error message so 
        # that only the second option is mentioned, which is the least 
        # confusing in this case.
        # 
        # It's worth noting that the real problem is that voluptuous checks 
        # default values.  I can't imagine a case where this is actually 
        # useful. I might think about making a PR for that.

        def do_check(x):
            if isinstance(x, MagicMock):
                return x
            return Schema(schema)(x)
        return do_check

    schema.update({
        Optional(k, default=MagicMock()): check(v)
        for k, v in expected.items()
    })
    return schema

# Something to think about: I'd like to put a version of this function in the 
# `parametrize_from_file` package.  I need a general way to specify the local 
# variables, though.  And to `eval()` the exception type...

def error(x):
    if x == 'none':
        return nullcontext()

    err_eval = eval_with(
            freezerbox=freezerbox,
            nx=nx,
            QueryError=freezerbox.QueryError,
            ParseError=freezerbox.ParseError,
            LoadError=freezerbox.LoadError,
    )
    if isinstance(x, str):
        err_type = err_eval(x)
        err_messages = []
    else:
        err_type = err_eval(x['type'])
        err_messages = x.get('message', [])
        if not isinstance(err_messages, list):
            err_messages = [err_messages]

    # Normally I'd use `@contextmanager` to make a context manager like this, 
    # but generator-based context managers cannot be reused.  This is a problem 
    # for tests, because if a test using this context manager is parametrized, 
    # the same context manager instance will need to be reused multiple times.  
    # The only way to support this is to implement the context manager from 
    # scratch.

    class expect_error:

        def __enter__(self):
            self.raises = pytest.raises(err_type)
            self.err = self.raises.__enter__()

        def __exit__(self, *args):
            if self.raises.__exit__(*args):
                for msg in err_messages:
                    self.err.match(msg)
                return True

    return expect_error()

def approx_Q(x):
    q = Quantity.from_anything(x)
    return pytest.approx(q, abs=Quantity(1e-6, q.unit))

eval_freezerbox = eval_with(
        freezerbox=freezerbox,
        nx=nx,
        Quantity=Quantity,
        Q=Q,
        approx_Q=approx_Q,
        MockReagent=MockReagent,
        MockMolecule=MockMolecule,
        MockNucleicAcid=MockNucleicAcid,
        TEST_DIR=TEST_DIR,
).all(freezerbox)
eval_pytest = eval_with().all(pytest)
eval_python = eval_with()


class Params:

    @classmethod
    def parametrize(cls, f):
        args = cls.args

        params = []
        for k, v in cls.__dict__.items():
            if k.startswith('params'):
                params.extend(v)

        # Could also check to make sure parameters make sense.

        return pytest.mark.parametrize(args, params)(f)




#!/usr/bin/env python3

"""\
Database of DNA sequences.
"""

__version__ = '0.18.0'

from .config import *
from .errors import *
from .model import *
from .fields import *
from .group_by import *
from .utils import *



#!/usr/bin/env python3

import sys

import stepwise
import freezerbox
import autoprop
import appcli
import pandas as pd

from stepwise import Quantity, pl, oxford_comma
from freezerbox import (
        MakerConfig, QueryError, convert_conc_unit, join_lists,
        iter_combo_makers, group_by_identity, unanimous,
)
from appcli import Key, Method, DocoptConfig
from more_itertools import all_equal
from dataclasses import dataclass
from pathlib import Path
from inform import warn, conjoin, plural

def parse_stocks(stock_strs):
    return join_lists(parse_stock(x) for x in stock_strs)

def parse_stock(stock_str):
    if stock_str:
        path = Path(stock_str)
        if path.exists():
            yield from load_stocks(path)
            return

    fields = stock_str.split(':')

    if len(fields) == 1:
        tag, = fields
        conc = None
        mw = None

    elif len(fields) == 2:
        tag, conc = fields
        conc = parse_conc(conc)
        mw = None

    elif len(fields) == 3:
        tag, conc, mw = fields
        conc = parse_conc(conc)
        mw = parse_mw(mw)

    else:
        raise ValueError(f"expected 1-3 colon-separated fields, got {len(fields)}: {stock_str!r}")

    yield Stock(tag=tag, conc=conc, mw=mw)

def load_stocks(path):
    try:
        # Nanodrop ONE
        nanodrop = pd.read_csv(path, sep='\t', encoding='utf-8', dtype=str)
    except UnicodeDecodeError:
        # Nanodrop 2000/2000c
        nanodrop = pd.read_csv(path, sep='\t', encoding='latin-1', dtype=str)

    tag_cols = 'Sample Name', 'Sample ID'
    conc_cols = 'Nucleic Acid(ng/uL)', 'Nucleic Acid'
    mw_cols = 'Molecular Weight', 'MW'

    def get_col(df, names, strict=True):
        for name in names:
            if name in df:
                return df[name]

        if strict:
            raise KeyError(names)
        else:
            return None

    df = pd.DataFrame()
    df['tag'] = get_col(nanodrop, tag_cols)
    df['conc_ng_uL'] = get_col(nanodrop, conc_cols)
    df['mw'] = get_col(nanodrop, mw_cols, strict=False)

    for i, row in df.iterrows():
        yield Stock(
                row.get('tag'),
                parse_conc(row.get('conc_ng_uL')),
                parse_mw(row.get('mw')),
        )

def parse_conc(conc_str):
    if not conc_str:
        return None

    try:
        conc = float(conc_str)
    except ValueError:
        pass
    else:
        return Quantity(conc, 'ng/µL')

    try:
        return Quantity.from_string(conc_str)
    except ValueError:
        pass

    raise ValueError(f"cannot parse {conc_str!r} as a concentration")

def parse_mw(mw_str):
    if not mw_str:
        return None

    return float(mw_str)

def parse_volume(vol_str):
    volumes = [
            float(v.strip())
            for v in vol_str.split(',')
    ]
    return volumes[0] if len(volumes) == 1 else volumes

@dataclass
class Stock:
    tag: str
    conc: Quantity
    mw: float

    def __init__(self, tag, conc=None, mw=None):
        self.tag = tag
        self.conc = conc
        self.mw = mw

@autoprop
class Dilute(appcli.App):
    """\
Calculate dilutions.

Usage:
    dilute <stocks>... (-v <µL> | -w <µL> | -V <µL>) [-c <conc>] [-C <conc>] 
        [-w <Da>] [-d <name>]

Arguments:
    <stocks>
        A list of stock solutions to dilute.  Any number of stocks can be 
        specified, and each argument can take one of two forms:

        Colon-separated fields:
            Specify the name and, if necessary, the concentration and molecular 
            weight for a single stock solution:

                <name>[:<conc>[:<mw>]]

            If either of the optional parameters aren't specified, they will be 
            looked up in the FreezerBox database using the given name.  Note 
            that the molecular weight is only required for certain unit 
            conversions, e.g. ng/µL to nM.  The name can be empty (or not in 
            the database) so long as enough information to calculate the 
            dilutions is supplied on the command-line.

        Path to existing TSV file:
            Read names, concentrations, and molecular weights for any number of 
            stock solutions from the given TSV file.  Information is read from 
            the following columns (any other columns are ignored):

            name:               "Sample Name" or "Sample ID"
            concentration:      "Nucleic Acid(ng/uL)" or "Nucleic Acid"
            molecular weight:   "Molecular Weight" or "MW"

            These column names are intended to match the files exported by 
            NanoDrop ONE and NanoDrop 2000 spectrophotometers, although it's 
            certainly possible to make these files yourself.

Options:
    -c --conc <conc>
        The final concentration achieve after dilution.  If not specified, 
        concentrations will be queried from the FreezerBox database.

    -C --stock-conc <conc>
        The stock concentration to use in the dilution calculations.  This 
        option will override any concentrations specified in the FreezerBox 
        database, but not any specified by the <stocks> argument.

    -v --volume <uL>
        The volume of concentrated stock solution to use in each dilution.  
        This can either be a single value or a comma-separated list of values.  
        If a single value is given, that value will be used for each dilution.  
        If multiple values are given, there must be exactly one value for each 
        stock solution, and the values will be associated with the stock 
        solutions in the order they were given.

    -D --diluent-volume <uL>
        The volume of diluent to use in each dilution.  This can either be a 
        single value or a comma-separated list of values, see the `--volume` 
        option for details.

    -V --total-volume <µL>
        The combined volume of stock and diluent to reach after each dilution.  
        This can either be a single value or a comma-separated list of values, 
        see the `--volume` option for details.

    --mw <Da>
        The molecular weight to use in the dilution calculations, e.g. if 
        converting between ng/µL and nM.  This option will override any 
        molecular weights specified in the FreezerBox database, but not any 
        specified by the <stocks> argument.

    -d --diluent <name>
        The name of the diluent to use.

Database:
    Dilution protocols can appear in the "Cleanup" column of a FreezerBox 
    database:

        dilute [conc=<conc>] [diluent=<name>]

    conc=<conc>
        See --conc.

        This setting is automatically applied as the concentration of the 
        associated database entry, unless superseded by the "Concentration" 
        column or a later cleanup step.  In other words, this concentration may 
        be used by any protocol that queries the FreezerBox database.

        Often, a concentration specified by this setting is regarded as the 
        "desired concentration", with the "actual concentration" being given in 
        the "Concentration" column in the event that the desired concentration 
        cannot be reached.

    diluent=<name>
        See --diluent.

"""
    __config__ = [
            DocoptConfig,
            MakerConfig,
    ]

    stocks = appcli.param(
            Key(DocoptConfig, '<stocks>', cast=parse_stocks),
            Method(lambda self: [Stock(x.tag, None, None) for x in self.products]),
    )
    target_conc = appcli.param(
            Key(DocoptConfig, '--conc'),
            Key(MakerConfig, 'conc'),
            cast=parse_conc,
            default=None,
    )
    stock_conc = appcli.param(
            Key(DocoptConfig, '--stock-conc'),
            cast=parse_conc,
            default=None,
    )
    stock_volume_uL = appcli.param(
            Key(DocoptConfig, '--volume'),
            cast=parse_volume,
            default=None,
    )
    diluent_volume_uL = appcli.param(
            Key(DocoptConfig, '--diluent-volume'),
            cast=parse_volume,
            default=None,
    )
    target_volume_uL = appcli.param(
            Key(DocoptConfig, '--total-volume'),
            cast=parse_volume,
            default=None,
    )
    mw = appcli.param(
            Key(DocoptConfig, '--mw'),
            cast=parse_mw,
            default=None,
    )
    diluent = appcli.param(
            Key(DocoptConfig, '--diluent'),
            Key(MakerConfig, 'diluent'),
            default=None,
    )
    show_stub = appcli.param(
            default=False,
    )

    def __bareinit__(self):
        self._db = None

    def __init__(self, stocks):
        self.stocks = stocks

    @classmethod
    def make(cls, db, products):

        def factory():
            app = cls.from_params()
            app.db = db
            app.show_stub = True
            return app

        yield from iter_combo_makers(
                factory,
                map(cls.from_product, products),
                group_by={
                    'target_conc': group_by_identity,
                    'diluent': group_by_identity,
                },
        )

    @classmethod
    def from_product(cls, product):
        app = cls.from_params()
        app.products = [product]
        app.load(MakerConfig)
        return app

    def get_db(self):
        if self._db is None:
            self._db = freezerbox.load_db()
        return self._db

    def set_db(self, db):
        self._db = db

    def get_concs(self):
        rows = []

        def first_valid(*options, strict=True, error=ValueError):
            for option in options:
                if callable(option):
                    try:
                        return option()
                    except QueryError:
                        continue

                if option is not None:
                    return option

            if strict:
                raise error
            else:
                return None

        for stock in self.stocks:
            rows.append(row := {})
            row['tag'] = tag = stock.tag
            row['target_conc'] = first_valid(
                    self.target_conc,
                    lambda: self.db[tag].conc,
                    error=ValueError(f"{tag}: no target concentration specified"),
            )
            row['stock_conc'] = first_valid(
                    stock.conc,
                    self.stock_conc,
                    lambda: self.db[tag].conc,
                    error=ValueError(f"{tag}: no stock concentration specified"),
            )
            row['mw'] = first_valid(
                    stock.mw,
                    self.mw,
                    lambda: self.db[tag].mw,
                    strict=False,
            )
            row['stock_conc_converted'] = convert_conc_unit(
                    row['stock_conc'],
                    row['mw'],
                    row['target_conc'].unit,
            )

        return pd.DataFrame(rows)

    def get_dilutions(self):
        self._check_volumes()

        df = self.concs
        k = df['stock_conc_converted'] / df['target_conc']
        dont_calc = (k <= 1)
        k[dont_calc] = pd.NA  # Avoid dividing by zero.

        if uL := self.stock_volume_uL:
            df['stock_uL'] = uL
            df['diluent_uL'] = uL * (k - 1)

        elif uL := self.diluent_volume_uL:
            df['diluent_uL'] = uL
            df['stock_uL'] = uL / (k - 1)
            uL = 'any'

        elif uL := self.target_volume_uL:
            df['stock_uL'] = uL / k
            df['diluent_uL'] = uL - df['stock_uL']

        df['final_conc'] = df['target_conc']
        df.loc[dont_calc, 'final_conc'] = df.loc[dont_calc, 'stock_conc_converted']
        df.loc[dont_calc, 'stock_uL'] = uL
        df.loc[dont_calc, 'diluent_uL'] = 0

        df['too_dilute'] = df['target_conc'] != df['final_conc']
        if df['too_dilute'].any():
            tags = df['tag'][df['too_dilute']]
            warn(f"the following {plural(tags):/stock is/stocks are} too dilute: {', '.join(tags)}")

        return df

    def get_protocol(self):
        if self.show_stub:
            return self.stub_protocol
        else:
            return self.full_protocol

    def get_stub_protocol(self):
        tags = [x.tag for x in self.stocks]
        in_diluent = f' in {self.diluent}' if self.diluent else ''

        p = stepwise.Protocol()
        p += f"Dilute {oxford_comma(tags)} to {self.target_conc}{in_diluent}."
        return p

    def get_full_protocol(self):
        df = self.dilutions
        uL = lambda x: x if isinstance(x, str) else f'{x:.2f} µL'

        ## Main table:

        main_header = [
                "Name",
                "Stock Vol",
                "Diluent Vol",
        ]
        main_row = lambda x: [
                x['tag'],
                uL(x['stock_uL']),
                uL(x['diluent_uL']),
        ]
        main_align = '<>>'

        to_conc = ""
        in_diluent = ""

        try:
            target_conc = unanimous(df['target_conc'])
        except ValueError:
            show_conc = True
        else:
            show_conc = False
            to_conc = f"to {target_conc} "

        if self.diluent:
            in_diluent = f"in {self.diluent} "

        if show_conc or df['too_dilute'].any():
            main_header.append("Final Conc")
            main_row_3 = main_row
            main_row = lambda x: [
                    *main_row_3(x),
                    str(x['final_conc']),
            ]
            main_align += '>'

        main_table = stepwise.table(
                rows=[main_row(x) for _, x in df.iterrows()],
                header=main_header,
                align=main_align,
        )

        ## Supplementary table:

        def stock_str(row):
            c0, c1 = row['stock_conc'], row['stock_conc_converted']
            return f'{c0} = {c1}' if c0.unit != c1.unit else f'{c1}'

        def mw_str(row):
            mw = row['mw']
            return f'{mw:.1f}' if mw else '?'

        supp_table = stepwise.table(
                rows=[
                    [x['tag'], mw_str(x), stock_str(x), x['target_conc']]
                    for _, x in df.iterrows()
                ],
                header=['Name', 'MW', 'Stock Conc', 'Target Conc'],
                align='<>>>',
        )

        ## Protocol:

        p = stepwise.Protocol()
        p += pl(
                f"Dilute the following {plural(df):stock solution/s} {to_conc}{in_diluent}[1]:",
                main_table,
        )
        p.footnotes[1] = pl(
                "Concentrations:",
                supp_table,
        )
        return p

    def get_product_conc(self):
        return self.target_conc

    def _check_volumes(self):
        volumes = [
                ('stock volume', self.stock_volume_uL),
                ('diluent volume', self.diluent_volume_uL),
                ('target volume', self.target_volume_uL),
        ]
        given = [k for k, v in volumes if v]
        if len(given) == 0:
            raise ValueError("no volumes specified")
        if len(given) > 1:
            raise ValueError(f"specified {' and '.join(given)}")

if __name__ == '__main__':
    app = Dilute.from_params()
    app.load(DocoptConfig)
    app.protocol.print()




#!/usr/bin/env python3

import pytest
import parametrize_from_file

from freezerbox import Tag
from freezerbox.model import *
from freezerbox.utils import *
from stepwise import Quantity
from schema_helpers import *
from pytest import approx

class PrefixParams(Params):
    args = 'args, expected'
    params = [
            # 'x' is from the various mock reagents.
            ([],                                    'xbrfpo'),
            ([Reagent],                             'xbrfpo'),
            ([Buffer],                              'b'),
            ([Protein],                             'r'),
            ([Protein, NucleicAcid],                'xrfpo'),
            ([NucleicAcid],                         'xfpo'),
            ([NucleicAcid, Plasmid],                'xfpo'),
            ([NucleicAcid, Oligo],                  'xfpo'),
            ([NucleicAcid, Plasmid, Oligo],         'xfpo'),
            ([Plasmid],                             'p'),
            ([Oligo],                               'o'),
            ([Plasmid, Oligo],                      'po'),
    ]

parse_schema = lambda input_name, schema={}: Schema({
    input_name: str,
    Optional('kwargs', default={}): eval_freezerbox,
    **error_or({
        'parsed': eval_freezerbox,
        'converted': {str: eval},
    }),
    **schema,
})

@parametrize_from_file
def test_normalize_seq(raw_seq, expected):
    assert normalize_seq(raw_seq) == expected

@PrefixParams.parametrize
def test_get_tag_prefixes(args, expected):
    assert get_tag_prefixes(*args) == set(expected)

@PrefixParams.parametrize
def test_get_tag_pattern(args, expected):
    assert re.fullmatch(
            fr'\[[{expected}]+\]\\d\+',
            get_tag_pattern(*args),
    )

@parametrize_from_file(
        schema=Schema({
            'tag_str': str,
            **error_or({
                'expected': {
                    'type': str,
                    'id': Coerce(int),
                },
            }),
        }),
)
def test_parse_tag(tag_str, expected, error):
    with error:
        assert parse_tag(tag_str) == Tag(**expected)

@parametrize_from_file(
        schema=Schema({
            'bool_str': str,
            **error_or({
                'expected': eval,
            }),
        }),
)
def test_parse_bool(bool_str, expected, error):
    with error:
        assert parse_bool(bool_str) == expected

@parametrize_from_file(schema=parse_schema('time_str'))
def test_parse_time(time_str, kwargs, parsed, converted, error):
    with error:
        assert parse_time(time_str, **kwargs) == parsed
    with error:
        assert parse_time_s(time_str, **kwargs) == approx(converted['s'])
    with error:
        assert parse_time_m(time_str, **kwargs) == approx(converted['m'])
    with error:
        assert parse_time_h(time_str, **kwargs) == approx(converted['h'])

@parametrize_from_file(
        schema=Schema({
            'given': eval,
            'expected': str,
        })
)
def test_format_time_s(given, expected):
    assert format_time_s(given) == expected

@parametrize_from_file(
        schema=Schema({
            'given': eval,
            'expected': str,
        })
)
def test_format_time_m(given, expected):
    assert format_time_m(given) == expected

@parametrize_from_file(schema=parse_schema('temp_str'))
def test_parse_temp(temp_str, kwargs, parsed, converted, error):
    with error:
        assert parse_temp(temp_str, **kwargs) == parsed
    with error:
        assert parse_temp_C(temp_str, **kwargs) == approx(converted['C'])

@parametrize_from_file(schema=parse_schema('vol_str'))
def test_parse_volume(vol_str, kwargs, parsed, converted, error):
    with error:
        assert parse_volume(vol_str, **kwargs) == parsed
    with error:
        assert parse_volume_uL(vol_str, **kwargs) == approx(converted['uL'])
    with error:
        assert parse_volume_mL(vol_str, **kwargs) == approx(converted['mL'])

@parametrize_from_file(schema=parse_schema('mass_str'))
def test_parse_mass(mass_str, kwargs, parsed, converted, error):
    with error:
        assert parse_mass(mass_str, **kwargs) == parsed
    with error:
        assert parse_mass_ug(mass_str, **kwargs) == approx(converted['ug'])
    with error:
        assert parse_mass_mg(mass_str, **kwargs) == approx(converted['mg'])

@parametrize_from_file(schema=parse_schema('conc_str', {'mw': eval}))
def test_parse_conc(conc_str, mw, kwargs, parsed, converted, error):
    from itertools import combinations_with_replacement

    converted['uM'] = converted['nM'] / 1000
    converted['ug/uL'] = converted['ng/uL'] / 1000

    with error:
        assert parse_conc(conc_str, **kwargs) == parsed

    with error:
        assert parse_conc_nM(conc_str, mw, **kwargs) == converted['nM']
    with error:
        assert parse_conc_uM(conc_str, mw, **kwargs) == converted['uM']
    with error:
        assert parse_conc_ng_uL(conc_str, mw, **kwargs) == converted['ng/uL']
    with error:
        assert parse_conc_ug_uL(conc_str, mw, **kwargs) == converted['ug/uL']

    for unit, value in converted.items():
        q_given = parsed
        mw_given = mw if unit != q_given.unit else None
        q_expected = Quantity(value, unit)
        q_converted = convert_conc_unit(q_given, mw_given, unit)
        assert q_converted == pytest.approx(
                q_expected,
                abs=Quantity(1e-6, unit),
        )

@parametrize_from_file(schema=parse_schema('size_str'))
def test_parse_size(size_str, kwargs, parsed, converted, error):
    with error:
        assert parse_size(size_str, **kwargs) == parsed
    with error:
        assert parse_size_bp(size_str, **kwargs) == converted['bp']
    with error:
        assert parse_size_kb(size_str, **kwargs) == converted['kb']

@parametrize_from_file(
        schema=Schema({
            'molecule': eval,
            Optional('default_strandedness', default='None'): eval,
            **error_or({
                'expected': eval,
            }),
        })
)
def test_parse_stranded_molecule(molecule, default_strandedness, expected, error):
    with error:
        actual = parse_stranded_molecule(molecule, default_strandedness)
        assert actual == expected

@parametrize_from_file(
        schema=Schema({
            'len': Coerce(int),
            'molecule': eval,
            'expected': Coerce(float),
        }),
)
def test_mw_from_length(len, molecule, expected):
    assert mw_from_length(len, molecule) == approx(expected)

@parametrize_from_file(
        schema=Schema({
            'items': eval,
            Optional('kwargs', default={}): {str: eval},
            **error_or({
                'expected': eval,
            }),
        }),
)
def test_unanimous(items, kwargs, expected, error):
    with error:
        assert unanimous(items, **kwargs) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_join_lists(given, expected):
    assert join_lists(given) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_join_dicts(given, expected):
    assert join_dicts(given) == expected

@parametrize_from_file(schema=Schema({str: eval}))
def test_join_sets(given, expected):
    assert join_sets(given) == expected




#!/usr/bin/env python3

import freezerbox
import parametrize_from_file

from stepwise.testing import disable_capture
from freezerbox import Database, parse_fields
from freezerbox.stepwise.make import Make
from more_itertools import one
from schema_helpers import *
from mock_model import *
from os import getcwd

def _db_from_maker(x):
    if 'maker' in x:
        x['db'] = db = Database()
        x['tag'] = tag = 'x1'

        db[tag] = MockMolecule(
                synthesis=parse_fields(x['maker']),
        )
        del x['maker']

    return x

def _guess_tag(x):
    if 'tag' not in x:
        x['tag'] = one(x['db'], too_short=Invalid, too_long=Invalid)
    return x


@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            Optional('tags', default=[]): [str],
            'expected': empty_ok([str]),
        }),
)
def test_make(db, tags, expected, disable_capture, mock_plugins):
    cwd = getcwd()

    tags = tags or list(db.keys())
    app = Make(db, tags)

    with disable_capture:
        assert app.protocol.steps == expected

    assert getcwd() == cwd

@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            'tags': [str],
            Optional('kwargs', default={}): {str: eval_pytest},
            'expected': [str],
        }),
)
def test_iter_targets(db, tags, kwargs, expected, mock_plugins):
    from freezerbox.stepwise.make import iter_targets
    targets = {str(x.tag) for x in iter_targets(db, tags, **kwargs)}
    assert targets == set(expected)

@parametrize_from_file(
        schema=Schema(
            Or(
                And(
                    {
                        'db': eval_db,
                        'tag': str,
                        Optional('maker_attrs', default={}): {
                            str: eval_freezerbox,
                        },
                        Optional('reagent_attrs', default={}): {
                            str: eval_freezerbox,
                        },
                    },
                    _guess_tag,
                ),
                And(
                    {
                        'maker': str,
                        Optional('maker_attrs', default={}): {
                            str: eval_freezerbox,
                        },
                        Optional('reagent_attrs', default={}): {
                            str: eval_freezerbox,
                        },
                    },
                    _db_from_maker,
                ),
            ),
        ),
)
def test_builtin_maker_attrs(db, tag, maker_attrs, reagent_attrs, disable_capture):
    with disable_capture:
        reagent = db[tag]

        for key, value in maker_attrs.items():
            i, attr = key.split('.'); i = int(i)
            maker = reagent.make_intermediate(i).maker
            assert getattr(maker, attr) == value

        for attr, value in reagent_attrs.items():
            assert getattr(reagent, attr) == value



#!/usr/bin/env python3

"""\
Query the DNA database.

Usage:
    freezerbox query <tags>... [--attrs <list>]
    freezerbox check [<db>]
    freezerbox which

Commands:
    query:  Display information about of the given constructs.
    check:  Make sure everything in the current database makes sense.
    which:  List the available databases.

Options:
    -a --attrs <list>      [default: length,mw]
        A comma-separated list of attributes to display.

Arguments:
    <tags>
        One or more tags to search for in the database.

    <db>
        The name of a database to use.  This corresponds to the "use" setting 
        in the configuration files.  If not given, the default database for the 
        current working directory will be used.
"""

import docopt
from tqdm import tqdm
from freezerbox import load_config, load_db, CheckError, QueryError

def main():
    args = docopt.docopt(__doc__)

    if args['query']:
        query(args['<tags>'], args['--attrs'].split(','))

    if args['check']:
        check(args['<db>'])

    if args['which']:
        which()

def query(tags, attrs):
    db = load_db()

    print('\t'.join(['tag', *attrs]))

    for tag in tags:
        try:
            hit = db[tag]
        except QueryError:
            print(tag)
            continue

        row = []
        for attr in attrs:
            try:
                row.append(getattr(hit, attr))
            except (QueryError, AttributeError):
                row.append('')

        print('\t'.join(str(x) for x in [hit.tag, *row]))

def check(use=None):
    db = load_db(use)
    n_total = len(db)
    n_passed = 0

    for tag, construct in tqdm(db.items()):
        try:
            construct.check()
            n_passed += 1
        except CheckError as err:
            err.report()

    print(f"{n_total} sequences found.")
    print(f"{n_passed} passed all checks.")

def which():
    config = load_config()

    if 'use' not in config:
        print("No default database.")
    else:
        print(f"Default database: {config['use']!r}")

    print()

    if 'database' not in config:
        print("No available databases.")
    else:
        print("Available databases:")
        for db in config['database']:
            print(f"  {db!r} ({config['database'][db]['type']})")






#!/usr/bin/env python3

import re, os
from stepwise import Quantity
from more_itertools import split_when
from contextlib import contextmanager
from inform import did_you_mean
from .errors import ParseError, only_raise

NO_DEFAULT = object()

TIME_CONVERSION_FACTORS = {
    's':        60*60,
    'sec':      60*60,
    'second':   60*60,
    'seconds':  60*60,
    'm':        60,
    'min':      60,
    'minute':   60,
    'minutes':  60,
    'h':        1,
    'hr':       1,
    'hour':     1,
    'hours':    1,
}
TEMP_CONVERSION_FACTORS = {
        '°C': 1,
         'C': 1,
}
VOLUME_CONVERSION_FACTORS = {
        'nL': 1e9,
        'uL': 1e6,
        'µL': 1e6,
        'mL': 1e3,
         'L': 1,
}
MASS_CONVERSION_FACTORS = {
        'pg': 1e12,
        'ng': 1e9,
        'ug': 1e6,
        'µg': 1e6,
        'mg': 1e3,
         'g': 1,
        'kg': 1e-3,
}
CONC_CONVERSION_FACTORS_MOLARITY = {
        'pM': 1e12,
        'nM': 1e9,
        'uM': 1e6,
        'µM': 1e6,
        'mM': 1e3,
         'M': 1,
}
CONC_CONVERSION_FACTORS_DENSITY = {
        'ng/uL': 1e3,
        'ng/µL': 1e3,
        'µg/mL': 1e3,
        'ug/mL': 1e3,
        'µg/µL': 1,
        'ug/uL': 1,
        'mg/mL': 1,
}
CONC_UNITS = {
        *CONC_CONVERSION_FACTORS_MOLARITY,
        *CONC_CONVERSION_FACTORS_DENSITY,
}
SIZE_CONVERSION_FACTORS = {
        'bp': 1e3,
        'kb': 1,
}

def normalize_seq(raw_seq):
    # Ignore nonstandard nucleotides; they're too hard to deal with properly.

    # This regular expression works because `re.sub()` only substitutes the 
    # left-most occurrence of any overlapping patterns.  The non-greedy * is 
    # necessary to avoid eliding everything between the first and last 
    # nonstandard nucleotide.
    seq = re.sub(r'/.*?/', 'X', raw_seq)

    return seq.upper()

def get_tag_prefixes(*args):
    from .model import Reagent

    def add(prefixes, cls):
        if cls.tag_prefix:
            prefixes.add(cls.tag_prefix)
        for subcls in cls.__subclasses__():
            add(prefixes, subcls)

    prefixes = set()
    for cls in args or [Reagent]:
        add(prefixes, cls)
    
    return prefixes

def get_tag_pattern(*args):
    prefix_chars = ''.join(sorted(get_tag_prefixes(*args)))
    return fr'[{prefix_chars}]\d+'

@only_raise(ParseError)
def parse_tag(tag_str):
    from .model import Tag, get_tag_prefixes

    if isinstance(tag_str, Tag):
        return tag_str
    if isinstance(tag_str, tuple):
        return Tag(*tag_str)

    pfo = get_tag_prefixes()
    if m := re.fullmatch(fr'\s*(?P<type>[{pfo}])(?P<id>\d+)\s*', tag_str):
        return Tag(m.group('type'), int(m.group('id')))
    else:
        raise ParseError(f"expected a tag (e.g. 'p100'), not {tag_str!r}")

@only_raise(ParseError)
def parse_bool(bool_str):
    if bool_str.lower() in ('1', '+', 'x', 'y', 'yes', 'true'):
        return True
    if bool_str.lower() in ('0', '-',  '', 'n', 'no', 'false'):
        return False

    raise ParseError(f"can't interpret {bool_str!r} as a bool")

@only_raise(ParseError)
def parse_time(time_str, default_unit=None):

    # First, try parsing the time using a custom syntax, e.g. '1m30':
    time_pattern = r'(\d+)([hm])(\d+)'
    if m := re.fullmatch(time_pattern, time_str):
        value = 60 * int(m.group(1)) + int(m.group(3))
        unit = {'h': 'm', 'm': 's'}[m.group(2)]
        return Quantity(value, unit)

    # If that doesn't work, try parsing the time normally, e.g. '1 min'
    return _parse_quantity(
            time_str, TIME_CONVERSION_FACTORS, default_unit, 'time')

@only_raise(ParseError)
def parse_time_s(time_str, default_unit=None):
    time = parse_time(time_str, default_unit)
    time = time.convert_unit('s', TIME_CONVERSION_FACTORS)
    return time.value

@only_raise(ParseError)
def parse_time_m(time_str, default_unit=None):
    time = parse_time(time_str, default_unit)
    time = time.convert_unit('m', TIME_CONVERSION_FACTORS)
    return time.value

@only_raise(ParseError)
def parse_time_h(time_str, default_unit=None):
    time = parse_time(time_str, default_unit)
    time = time.convert_unit('h', TIME_CONVERSION_FACTORS)
    return time.value

def format_time_s(x):
    if x < 60:
        return f'{x}s'

    min = x // 60
    sec = x % 60

    return f'{min}m{f"{sec:02}" if sec else ""}'

def format_time_m(x):
    if x < 60:
        return f'{x}m'

    hr = x // 60
    min = x % 60

    return f'{hr}h{f"{min:02}" if min else ""}'

@only_raise(ParseError)
def parse_temp(temp_str, default_unit=None):
    return _parse_quantity(
            temp_str, TEMP_CONVERSION_FACTORS, default_unit, 'temperature')

@only_raise(ParseError)
def parse_temp_C(temp_str, default_unit=None):
    return _parse_and_convert_quantity(
            temp_str, TEMP_CONVERSION_FACTORS, default_unit, '°C', 'temperature')

@only_raise(ParseError)
def parse_volume(vol_str, default_unit=None):
    return _parse_quantity(
            vol_str, VOLUME_CONVERSION_FACTORS, default_unit, 'volume')

@only_raise(ParseError)
def parse_volume_uL(vol_str, default_unit=None):
    return _parse_and_convert_quantity(
            vol_str, VOLUME_CONVERSION_FACTORS, default_unit, 'µL', 'volume')

@only_raise(ParseError)
def parse_volume_mL(vol_str, default_unit=None):
    return _parse_and_convert_quantity(
            vol_str, VOLUME_CONVERSION_FACTORS, default_unit, 'mL', 'volume')

@only_raise(ParseError)
def parse_mass(mass_str, default_unit=None):
    return _parse_quantity(
            mass_str, MASS_CONVERSION_FACTORS, default_unit, 'mass')

@only_raise(ParseError)
def parse_mass_ug(mass_str, default_unit=None):
    return _parse_and_convert_quantity(
            mass_str, MASS_CONVERSION_FACTORS, default_unit, 'µg', 'mass')

@only_raise(ParseError)
def parse_mass_mg(mass_str, default_unit=None):
    return _parse_and_convert_quantity(
            mass_str, MASS_CONVERSION_FACTORS, default_unit, 'mg', 'mass')

@only_raise(ParseError)
def parse_conc(conc_str, default_unit=None):
    return _parse_quantity(conc_str, CONC_UNITS, default_unit, 'concentration')

@only_raise(ParseError)
def parse_conc_nM(conc_str, mw, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'nM')
    return conc.value

@only_raise(ParseError)
def parse_conc_uM(conc_str, mw, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'µM')
    return conc.value

@only_raise(ParseError)
def parse_conc_ng_uL(conc_str, mw, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'ng/µL')
    return conc.value

@only_raise(ParseError)
def parse_conc_ug_uL(conc_str, mw, default_unit=None):
    conc = parse_conc(conc_str, default_unit)
    conc = convert_conc_unit(conc, mw, 'µg/µL')
    return conc.value

@only_raise(ParseError)
def convert_conc_unit(conc, mw, new_unit):
    def pick_conversion_factors():
        if mw is not None:
            return {
                    **CONC_CONVERSION_FACTORS_MOLARITY,
                    **{
                        k: mw * v
                        for k, v in CONC_CONVERSION_FACTORS_DENSITY.items()
                    },
            }

        if conc.unit in CONC_CONVERSION_FACTORS_MOLARITY:
            return CONC_CONVERSION_FACTORS_MOLARITY

        if conc.unit in CONC_CONVERSION_FACTORS_DENSITY:
            return CONC_CONVERSION_FACTORS_DENSITY

        return {}

    return conc.convert_unit(new_unit, pick_conversion_factors())

@only_raise(ParseError)
def parse_size(size_str, default_unit=None):
    return _parse_quantity(size_str, SIZE_CONVERSION_FACTORS, default_unit, 'size')

@only_raise(ParseError)
def parse_size_bp(size_str, default_unit=None):
    return _parse_and_convert_quantity(
            size_str, SIZE_CONVERSION_FACTORS, default_unit, 'bp', 'size')

@only_raise(ParseError)
def parse_size_kb(size_str, default_unit=None):
    return _parse_and_convert_quantity(
            size_str, SIZE_CONVERSION_FACTORS, default_unit, 'kb', 'size')

def parse_stranded_molecule(molecule, default_strandedness=None):
    """
    Parse the molecule type (e.g. suitable input for Biopython functions) 
    and strandedness from a molecule string.  
    """
    if isinstance(molecule, tuple):
        return molecule

    stranded_molecules = {
            'dna':   (0, 'DNA'),
            'ssdna': (1, 'DNA'),
            'dsdna': (2, 'DNA'),
            'rna':   (0, 'RNA'),
            'ssrna': (1, 'RNA'),
            'dsrna': (2, 'RNA'),
    }
    try:
        strandedness, molecule = stranded_molecules[molecule.lower()]
    except KeyError:
        err = ParseError(molecule=molecule)
        err.brief = "unknown molecule type: {molecule!r}"
        err.info += f"expected: {', '.join(map(repr, stranded_molecules))}"
        raise err

    if not strandedness:
        strandedness = default_strandedness

    if not strandedness:
        strandedness = {'DNA': 2, 'RNA': 1}[molecule]

    return molecule, strandedness

def mw_from_length(len, molecule='dna'):
    molecule, strandedness = parse_stranded_molecule(molecule)

    # Parameters from: 
    # https://www.thermofisher.com/us/en/home/references/ambion-tech-support/rna-tools-and-calculators/dna-and-rna-molecular-weights-and-conversions.html
    params = {
            'DNA': (303.7, 79),
            'RNA': (320.5, 159),
    }
    m, b = params[molecule]
    return (m * len + b) * strandedness

def unanimous(
        items,
        default=NO_DEFAULT,
        err_empty=ValueError("empty iterable"),
        err_multiple=lambda v1, v2: ValueError(f"found multiple values: {v1!r}, {v2!r}"),
    ):
    it = iter(items)

    try:
        value = next(it)
    except StopIteration:
        if default is not NO_DEFAULT:
            return default
        else:
            raise err_empty

    for next_value in it:
        if next_value != value:
            raise err_multiple(value, next_value)

    return value

def join_lists(x):
    from itertools import chain
    return list(chain(*x))

def join_dicts(x):
    from collections import ChainMap
    return dict(ChainMap(*x))

def join_sets(x):
    return set.union(set(), *x)

@contextmanager
def cd(dir):
    try:
        prev_cwd = os.getcwd()
        os.chdir(dir)
        yield

    finally:
        os.chdir(prev_cwd)


def _parse_quantity(quantity_str, expected_units, default_unit, unit_type):
    try:
        if default_unit:
            quantity = Quantity.from_string_or_float(quantity_str, default_unit)
        else:
            quantity = Quantity.from_string(quantity_str)

    except ValueError:
        raise ParseError(
                "can't interpret {quantity_str!r} as a {unit_type}",
                quantity_str=quantity_str,
                unit_type=unit_type,
        ) from None

    if quantity.unit not in expected_units:
        raise ParseError(
                lambda e: f"can't interpret {e.quantity_str!r} as a {unit_type}, did you mean '{Quantity(e.quantity.value, did_you_mean(e.quantity.unit, e.expected_units))}'?",
                quantity=quantity,
                quantity_str=quantity_str,
                expected_units=expected_units,
                unit_type=unit_type,
        )

    return quantity

def _parse_and_convert_quantity(quantity_str, conversion_factors, default_unit, desired_unit, unit_type):
    quantity = _parse_quantity(
            quantity_str, conversion_factors, default_unit, unit_type)
    converted = quantity.convert_unit(desired_unit, conversion_factors)
    return converted.value





#!/usr/bin/env python3

import freezerbox
import networkx as nx
from schema_helpers import *
from mock_model import *

@parametrize_from_file(
        schema=Schema({
            'nodes': empty_ok({eval: eval}),
            'edges': empty_ok([And(eval, tuple)]),
            **error_or({
                'expected': empty_ok([And(eval, tuple)]),
            }),
        }),
)
def test_grouped_topological_sort(nodes, edges, expected, error):
    g = nx.DiGraph()
    g.add_edges_from(edges)

    for node, group in nodes.items():
        g.add_node(node, group=group)

    with error:
        actual = freezerbox.grouped_topological_sort(g)
        assert actual == expected

@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            'expected': empty_ok([{'arg0': str, 'tags': str}]),
        }),
)
def test_group_by_synthesis(db, expected, mock_plugins):
    actual = [
            (k, [str(v.tag) for v in vs])
            for k, vs in freezerbox.group_by_synthesis(db.values())
    ]
    expected = [
            (x['arg0'], x['tags'].split())
            for x in expected
    ]

    assert actual == expected

@parametrize_from_file(
        schema=Schema({
            'db': eval_db,
            'expected': empty_ok([{'arg0': str, 'ids': str}]),
        }),
)
def test_group_by_cleanup(db, expected, mock_plugins):
    actual = [
            (k, [v.maker_args['id'] for v in vs])
            for k, vs in freezerbox.group_by_cleanup(db.values())
    ]
    expected = [
            (x['arg0'], list(map(int, x['ids'].split())))
            for x in expected
    ]

    assert actual == expected
    








#!/usr/bin/env python3

import networkx as nx
from math import inf
from copy import deepcopy
from more_itertools import pairwise
from ..utils import parse_tag

def group_by_synthesis(products):
    products = list(products)

    # Sort groups by order of appearance:

    group_from_arg0 = {}
    arg0_from_group = {}
    next_group = 0

    for product in products:
        arg0 = product.synthesis_args.by_index[0]
        if arg0 not in arg0_from_group:
            group_from_arg0[arg0] = next_group
            arg0_from_group[next_group] = arg0
            next_group += 1

    # Construct a dependency graph:

    deps = nx.DiGraph()

    for i, product in enumerate(products):
        arg0 = product.synthesis_args.by_index[0]
        deps.add_node(
                product.tag,
                group=group_from_arg0[arg0],
                order=i,
        )

    for product in products:
        for dep in product.dependencies:
            dep = parse_tag(dep)
            if dep in deps:
                deps.add_edge(dep, product.tag)

    # Split into groups and yield intermediates:

    intermediate_from_tag = {
            x.tag: x.make_intermediate(0)
            for x in products
    }

    for group, tags in grouped_topological_sort(deps):
        arg0 = arg0_from_group[group]
        intermediates = [intermediate_from_tag[tag] for tag in tags]
        yield arg0, intermediates

def group_by_cleanup(products):
    products = list(products)

    # Construct a dependency graph:

    deps = nx.DiGraph()

    for i, product in enumerate(products):
        for j, cleanup in enumerate(product.cleanup_args):
            deps.add_node((i, j), group=cleanup.by_index[0])

        for pair in pairwise(enumerate(product.cleanup_args)):
            (j, args_j), (k, args_k) = pair
            deps.add_edge((i, j), (i, k))

    # Split into groups:

    for key, nodes in grouped_topological_sort(deps):
        intermediates = [
                products[i].make_intermediate(j+1)
                for i, j in nodes
        ]
        yield key, intermediates

def grouped_topological_sort(deps):
    """
    Arguments:
        deps: networkx.DiGraph
            A graph of the dependencies to account for.  Each node should have 
            a "group" attribute identifying which group it is part of.  The 
            returned groups will be sorted by this attribute when possible.
    """

    by_order = lambda x: deps.nodes[x].get('order', x)

    def inner_sort(candidates, dep_counts):
        best_groups = []
        best_score = (inf, inf)

        for type in candidates:
            next_candidates = deepcopy(candidates)
            next_dep_counts = deepcopy(dep_counts)
            next_group = next_candidates.pop(type)

            for node in next_group:
                for _, child in deps.edges(node):
                    next_dep_counts[child] -= 1
                    if next_dep_counts[child] == 0:
                        child_type = deps.nodes[child]['group']
                        next_candidates.setdefault(child_type, []).append(child)
                        del next_dep_counts[child]

            remaining_groups = inner_sort(
                    next_candidates,
                    next_dep_counts,
            )
            score = len(remaining_groups), type

            if score < best_score:
                best_score = score
                best_groups = [
                        (type, sorted(next_group, key=by_order)),
                        *remaining_groups,
                ]

        if dep_counts and not candidates:
            raise nx.NetworkXUnfeasible("graph contains a cycle")

        return best_groups

    candidates = {}
    dep_counts = {}

    for v, d in deps.in_degree():
        if d > 0:
            dep_counts[v] = d
        else:
            type = deps.nodes[v]['group']
            candidates.setdefault(type, []).append(v)

    return inner_sort(candidates, dep_counts)




#!/usr/bin/env python3

from schema_helpers import *

class DummyItem:

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __repr__(self):
        kwargs = ', '.join(f'{k}={v!r}' for k,v in self.kwargs.items())
        return f'DummyItem({kwargs})'

    def __eq__(self, other):
        return self.kwargs == other.kwargs

    def __getattr__(self, name):
        return self.kwargs[name]


@parametrize_from_file(
        schema=Schema({
            'grouper': eval_freezerbox,
            'items': eval,
            Optional('key', default='lambda x: x'): eval,
            'expected': empty_ok([{
                'value': eval_pytest,
                'items': eval_pytest,
            }]),
        }),
)
def test_group_by(grouper, items, key, expected):
    actual = [
            (k, list(it))
            for k, it in grouper(items, key=key)
    ]
    expected = [
            (x['value'], list(x['items']))
            for x in expected
    ]
    assert actual == expected

@parametrize_from_file(
        schema=Schema({
            'items': empty_ok([eval_with(DummyItem=DummyItem)]),
            'group_by': empty_ok({str: eval_freezerbox}),
            'merge_by': empty_ok({str: eval_freezerbox}),
            Optional('keys', default={}): empty_ok({str: eval_freezerbox}),
            'expected': empty_ok([{
                'attrs': empty_ok({str: eval_pytest}),
                'items': [eval_with(DummyItem=DummyItem)],
            }]),
        }),
)
def test_iter_combos(items, group_by, merge_by, keys, expected):
    actual = [
            (attrs, list(items))
            for attrs, items in freezerbox.iter_combos(
                items,
                group_by=group_by,
                merge_by=merge_by,
                keys=keys,
            )
    ]
    expected = [
            (x['attrs'], list(x['items']))
            for x in expected
    ]
    assert actual == expected



#!/usr/bin/env python3

import freezerbox
import pytest
import datetime
import autoprop
from freezerbox import Fields, QueryError
from freezerbox.model import *
from schema_helpers import *
from mock_model import *

kwargs_schema = lambda expected=eval_freezerbox: Schema({
    Optional('kwargs', {}): empty_ok({str: eval_freezerbox}),
    **error_or({
        'expected': expected,
    }),
})

def test_database_getitem_setitem():
    db = Database()

    x0 = MockReagent()
    with pytest.raises(QueryError, match='not attached'): x0.db
    with pytest.raises(QueryError, match='not attached'): x0.tag
    with pytest.raises(QueryError, match="x0: not found in database"): db['x0']

    db['x1'] = x1 = MockReagent()
    db['x02'] = x2 = MockReagent()
    db['x', 3] = x3 = MockReagent()
    db[Tag('x', 4)] = x4 = MockReagent()

    with pytest.raises(LoadError, match="x1: already in database, cannot be replaced"):
        db['x1'] = MockReagent()
    with pytest.raises(LoadError, match="MockReagent\\(\\) cannot have tag 'p1'\n. expected 'x' prefix"):
        db['p1'] = MockReagent()

    assert x1.db is db
    assert x2.db is db
    assert x3.db is db
    assert x4.db is db

    assert x1.tag == Tag('x', 1)
    assert x2.tag == Tag('x', 2)
    assert x3.tag == Tag('x', 3)
    assert x4.tag == Tag('x', 4)

    assert db['x1'] is x1
    assert db['x01'] is x1
    assert db['x', 1] is x1
    assert db[Tag('x', 1)] is x1

    assert db['x2'] is x2
    assert db['x02'] is x2
    assert db['x', 2] is x2
    assert db[Tag('x', 2)] is x2

    assert db['x3'] is x3
    assert db['x03'] is x3
    assert db['x', 3] is x3
    assert db[Tag('x', 3)] is x3

    assert db['x4'] is x4
    assert db['x04'] is x4
    assert db['x', 4] is x4
    assert db[Tag('x', 4)] is x4

def test_database_delitem():
    db = Database()
    db['x1'] = x1 = MockReagent()

    assert x1 in db
    assert x1.db is db
    assert x1.tag == Tag('x', 1)

    del db['x1']

    assert x1 not in db
    with pytest.raises(QueryError, match='not attached'): x1.db
    with pytest.raises(QueryError, match='not attached'): x1.tag

def test_database_contains():
    db = Database()
    x1 = MockReagent()
    assert x1 not in db

    db['x1'] = x1
    assert x1 in db

def test_database_iter():
    db = Database()
    db['x1'] = x1 = MockReagent()
    db['x2'] = x2 = MockReagent()

    values = {x1, x2}
    keys = {x.tag for x in values}
    items = {(x.tag, x) for x in values}

    assert set(db) == keys
    assert set(db.keys()) == keys
    assert set(db.values()) == values
    assert set(db.items()) == items

def test_database_len():
    db = Database()
    assert len(db) == 0

    db['x1'] = MockReagent()
    assert len(db) == 1

    db['x2'] = MockReagent()
    assert len(db) == 2


def test_tag():
    tag = Tag('p', 1)

    assert tag.type == 'p'
    assert tag.id == 1
    assert str(tag) == 'p1'


def test_reagent_repr():
    x1 = MockReagent()
    x2 = MockReagent(a='b')
    x3 = MockReagent(a='b', c='d')

    assert repr(x1) == "MockReagent()"
    assert repr(x2) == "MockReagent(a='b')"
    assert repr(x3) == "MockReagent(a='b', c='d')"

    db = Database()
    db['x1'] = x1
    db['x2'] = x2
    db['x3'] = x3

    assert repr(x1) == "MockReagent('x1')"
    assert repr(x2) == "MockReagent('x2', a='b')"
    assert repr(x3) == "MockReagent('x3', a='b', c='d')"

def test_reagent_intermediate_repr():
    x1 = MockReagent(
            synthesis=Fields(['m'], {}),
    )
    i1 = x1.make_intermediate(0)

    assert repr(x1) == "MockReagent(synthesis=Fields(['m'], {}))"
    assert repr(i1) == "MockReagentIntermediate(step=0, synthesis=Fields(['m'], {}))"

@parametrize_from_file(schema=kwargs_schema())
def test_reagent_name(kwargs, expected, error):
    db = Database()
    db['x1'] = x1 = MockReagent(**kwargs)
    with error:
        assert x1.name == expected

@parametrize_from_file(schema=kwargs_schema())
def test_reagent_alt_names(kwargs, expected, error):
    db = Database()
    db['x1'] = x1 = MockReagent(**kwargs)
    with error:
        assert x1.alt_names == expected

@parametrize_from_file(schema=kwargs_schema())
def test_reagent_date(kwargs, expected, error):
    db = Database()
    db['x1'] = x1 = MockReagent(**kwargs)
    with error:
        assert x1.date == expected

@parametrize_from_file(schema=kwargs_schema())
def test_reagent_desc(kwargs, expected, error):
    db = Database()
    db['x1'] = x1 = MockReagent(**kwargs)
    with error:
        assert x1.desc == expected

@parametrize_from_file(schema=kwargs_schema())
def test_reagent_ready(kwargs, expected, error):
    db = Database()
    db['x1'] = x1 = MockReagent(**kwargs)
    with error:
        assert x1.ready == expected

def test_reagent_maker_args_1():
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['m'], {'conc': '1 nM'}),
            cleanups=[
                Fields(['m'], {'conc': '2 nM'}),
                Fields(['m'], {'conc': '3 nM'}),
            ],
    )
    assert x1.synthesis_args[0] == 'm'
    assert x1.synthesis_args['conc'] == '1 nM'

    assert x1.cleanup_args[0][0] == 'm'
    assert x1.cleanup_args[0]['conc'] == '2 nM'

    assert x1.cleanup_args[1][0] == 'm'
    assert x1.cleanup_args[1]['conc'] == '3 nM'

def test_reagent_maker_args_2():
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=lambda: Fields(['m'], {'conc': '1 nM'}),
            cleanups=lambda: [
                Fields(['m'], {'conc': '2 nM'}),
                Fields(['m'], {'conc': '3 nM'}),
            ],
    )
    assert x1.synthesis_args[0] == 'm'
    assert x1.synthesis_args['conc'] == '1 nM'

    assert x1.cleanup_args[0][0] == 'm'
    assert x1.cleanup_args[0]['conc'] == '2 nM'

    assert x1.cleanup_args[1][0] == 'm'
    assert x1.cleanup_args[1]['conc'] == '3 nM'

def test_reagent_makers(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['m'], {'conc': '1 nM'}),
            cleanups=[
                Fields(['m'], {'conc': '2 nM'}),
                Fields(['m'], {'conc': '3 nM'}),
            ],
    )
    m1 = x1.synthesis_maker
    m2 = x1.cleanup_makers[0]
    m3 = x1.cleanup_makers[1]

    assert isinstance(m1, MockMaker)
    assert m1.product_conc == Quantity(1, 'nM')

    assert isinstance(m2, MockMaker)
    assert m2.product_conc == Quantity(2, 'nM')

    assert isinstance(m3, MockMaker)
    assert m3.product_conc == Quantity(3, 'nM')

def test_reagent_maker_err_1():
    db = Database()
    db['x1'] = x1 = MockReagent()
    with pytest.raises(QueryError, match="no synthesis specified"):
        x1.synthesis_args
    with pytest.raises(QueryError, match="no synthesis specified"):
        x1.synthesis_maker
    with pytest.raises(QueryError, match="no synthesis specified"):
        x1.cleanup_args
    with pytest.raises(QueryError, match="no synthesis specified"):
        x1.cleanup_makers

def test_reagent_maker_err_2(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockReagent(
            # no synthesis
            cleanups=[
                Fields(['m'], {'conc': '2 nM'}),
                Fields(['m'], {'conc': '3 nM'}),
            ],
    )
    with pytest.raises(QueryError, match="no synthesis specified"):
        x1.cleanup_args
    with pytest.raises(QueryError, match="no synthesis specified"):
        x1.cleanup_makers

def test_reagent_get_maker_attr_1(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['m'], {'conc': '1 nM'}),
            cleanups=[
                Fields(['m'], {'conc': '2 nM'}),
                Fields(['m'], {'conc': '3 nM'}),
            ],
    )
    assert x1.get_maker_attr('product_conc') == Quantity(3, 'nM')
    assert x1.get_maker_attr('product_conc', None) == Quantity(3, 'nM')

def test_reagent_get_maker_attr_2(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['m'], {'conc': '1 nM'}),
            cleanups=[
                Fields(['m'], {'conc': '2 nM'}),
                Fields(['m'], {}),
            ],
    )
    assert x1.get_maker_attr('product_conc') == Quantity(2, 'nM')
    assert x1.get_maker_attr('product_conc', None) == Quantity(2, 'nM')

def test_reagent_get_maker_attr_3(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['m'], {'conc': '1 nM'}),
            cleanups=[
                Fields(['m'], {}),
                Fields(['m'], {}),
            ],
    )
    assert x1.get_maker_attr('product_conc') == Quantity(1, 'nM')
    assert x1.get_maker_attr('product_conc', None) == Quantity(1, 'nM')

def test_reagent_get_maker_attr_4(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockReagent(
            synthesis=Fields(['m'], {}),
            cleanups=[
                Fields(['m'], {}),
                Fields(['m'], {}),
            ],
    )
    with pytest.raises(QueryError, match='product_conc'):
        x1.get_maker_attr('product_conc')
    assert x1.get_maker_attr('product_conc', None) == None

def test_reagent_get_maker_attr_5():
    db = Database()
    db['x1'] = x1 = MockReagent()
    with pytest.raises(QueryError, match='product_conc'):
        x1.get_maker_attr('product_conc')
    assert x1.get_maker_attr('product_conc', None) == None


@parametrize_from_file(schema=kwargs_schema())
def test_molecule_seq(kwargs, expected, error, mock_plugins):
    db = Database()
    db['x1'] = x1 = MockMolecule(**kwargs)
    with error:
        assert x1.seq == expected

        # Make sure the sequence is cached:
        x1._attrs['seq'] = '!!!'
        assert x1.seq == expected

@parametrize_from_file(schema=kwargs_schema())
def test_molecule_length(kwargs, expected, error):
    db = Database()
    db['x1'] = x1 = MockMolecule(**kwargs)
    with error:
        assert x1.length == expected

@parametrize_from_file(
        schema=kwargs_schema([eval_freezerbox]),
)
def test_molecule_conc(kwargs, expected, error, mock_plugins):
    db = Database()
    db['x1'] = x1 = MockMolecule(**kwargs)

    get_by_unit = {
            'nM': lambda: x1.conc_nM,
            'uM': lambda: x1.conc_uM,
            'µM': lambda: x1.conc_uM,
            'ng/uL': lambda: x1.conc_ng_uL,
            'ng/µL': lambda: x1.conc_ng_uL,
            'mg/mL': lambda: x1.conc_mg_mL,
    }

    # The first expected value should have the same units as the specified 
    # concentration.
    with error:
        assert x1.conc == expected[0]

    for q in expected:
        with error:
            assert x1.get_conc(q.unit).value == pytest.approx(q.value)
            assert x1.get_conc(q.unit).unit == q.unit
        with error:
            assert get_by_unit[q.unit]() == pytest.approx(q.value)

@parametrize_from_file(
        schema=kwargs_schema({str: eval_freezerbox}),
)
def test_molecule_volume(kwargs, expected, error, mock_plugins):
    db = Database()
    db['x1'] = x1 = MockMolecule(**kwargs)

    with error:
        assert x1.volume == expected['quantity']
    with error:
        assert x1.volume_uL == expected['uL']


def test_protein_mw():
    db = Database()
    db['r1'] = r1 = Protein(seq='DYKDDDDK')
    assert r1.mw == pytest.approx(1012.98, abs=0.1)

def test_protein_mw_err():
    db = Database()
    db['r1'] = r1 = Protein(seq='X')
    with pytest.raises(QueryError, match="'X' is not a valid unambiguous letter for protein"):
        r1.mw

@parametrize_from_file(
        schema=kwargs_schema({
            'molecule': eval,
            'strandedness': eval,
        }),
)
def test_nucleic_acid_molecule(kwargs, expected, error, mock_plugins):
    db = Database()
    db['f1'] = f1 = NucleicAcid(**kwargs)
    with error:
        assert f1.molecule == expected['molecule']
        assert f1.is_double_stranded == (expected['strandedness'] == 2)
        assert f1.is_single_stranded == (expected['strandedness'] == 1)

@parametrize_from_file(schema=kwargs_schema())
def test_nucleic_acid_circular(kwargs, expected, error, mock_plugins):
    db = Database()
    db['f1'] = f1 = NucleicAcid(**kwargs)
    with error:
        assert f1.is_circular == expected
        assert f1.is_linear == (not expected)

@parametrize_from_file(
        schema=kwargs_schema(),
)
def test_nucleic_acid_mw(kwargs, expected, error):
    # 5'-phosphorylation assumed.
    # http://molbiotools.com/dnacalculator.html
    db = Database()
    db['f1'] = f1 = NucleicAcid(**kwargs)
    with error:
        assert f1.mw == pytest.approx(expected, abs=0.1)

def test_plasmid_mw():
    # http://molbiotools.com/dnacalculator.html
    db = Database()
    db['p1'] = p1 = Plasmid(seq='ATCG')

    assert p1.is_circular == (not p1.is_linear) == True
    assert p1.is_double_stranded == (not p1.is_single_stranded) == True
    assert p1.mw == pytest.approx(2471.58, abs=0.1)

def test_oligo_mw():
    # 5'-OH assumed.
    # http://molbiotools.com/dnacalculator.html
    db = Database()
    db['o1'] = o1 = Oligo(seq='ATCG')

    assert o1.is_circular == (not o1.is_linear) == False
    assert o1.is_double_stranded == (not o1.is_single_stranded) == False
    assert o1.mw == pytest.approx(1173.82, abs=0.1)

@parametrize_from_file(schema=kwargs_schema())
def test_oligo_tm(kwargs, expected, error):
    db = Database()
    db['o1'] = o1 = Oligo(**kwargs)
    with error:
        assert o1.tm == pytest.approx(expected)


def test_intermediate(mock_plugins):
    db = Database()
    db['x1'] = x1 = MockMolecule(
            synthesis=Fields('m', {'conc': '1 nM'}),
            cleanups=[
                Fields('m', {'conc': '2 nM'}),
                Fields('m', {'conc': '3 nM'}),
            ],
    )

    # Access the concentration before creating the intermediates so that a 
    # concentration value is cached.  The intermediates will need to forget 
    # this value:
    assert x1.conc == Quantity(3, 'nM')

    i1 = x1.make_intermediate(0)
    i2 = x1.make_intermediate(1)
    i3 = x1.make_intermediate(2)

    with pytest.raises(QueryError) as err:
        x1.make_intermediate(3)

    assert err.match("x1: intermediate 3 doesn't exist")
    assert err.match("intermediates:")
    assert err.match("0: m conc='1 nM'")
    assert err.match("1: m conc='2 nM'")
    assert err.match("2: m conc='3 nM'")

    # Access the concentration again after creating the intermediates, to make 
    # sure that doing so does not somehow re-establish the cache:
    assert x1.conc == Quantity(3, 'nM')

    # Test that precursors are correct:
    assert i2.precursor is i1
    assert i3.precursor is i2

    # Test that fields are correct:
    assert i1.maker_args[0] == 'm'
    assert i1.maker_args['conc'] == '1 nM'
    assert i2.maker_args[0] == 'm'
    assert i2.maker_args['conc'] == '2 nM'
    assert i3.maker_args[0] == 'm'
    assert i3.maker_args['conc'] == '3 nM'

    # Test that makers are correct:
    assert i1.maker.product_conc == Quantity(1, 'nM')
    assert i2.maker.product_conc == Quantity(2, 'nM')
    assert i3.maker.product_conc == Quantity(3, 'nM')

    # Test that properties depending on fields are re-evaluated correctly.
    assert i1.conc == Quantity(1, 'nM')
    assert i2.conc == Quantity(2, 'nM')
    assert i3.conc == Quantity(3, 'nM')




#!/usr/bin/env python3

import appdirs
import appcli
import autoprop

from pathlib import Path
from configurator import Config
from functools import lru_cache
from appcli import unbind_method
from more_itertools import one, always_iterable
from operator import attrgetter

from .utils import parse_tag
from .errors import QueryError, ParseError, only_raise

@lru_cache
def load_config():
    config = Config()
    cwd = Path.cwd()
    dirs = [cwd, *cwd.parents]

    for dir in reversed(dirs):
        paths = [
                dir / '.config' / 'freezerbox' / 'conf.toml',
                dir / '.freezerboxrc',
        ]
        for path in paths:
            if path.exists():
                subconf = Config.from_path(path, parser='toml')
                config.merge(subconf)

        dir = dir.parent

    return config.data

@autoprop
class ReagentConfig(appcli.Config):
    autoload = True
    autoload_db = True
    db_getter = attrgetter('db')
    tag_getter = attrgetter('tag')
    transform = None

    @autoprop
    class Layer(appcli.Layer):

        def __init__(self, *, db_getter, tag_getter, transform_func, autoload_db):
            self.db_getter = db_getter
            self.tag_getter = tag_getter
            self.transform_func = transform_func
            self.autoload_db = autoload_db
            self.db = None

        def iter_values(self, key, log):
            # First: See if the object has a database attribute.  If it does, 
            # it costs nothing to access it, and it will allow us to log the 
            # path to the database before any subsequent steps fail.

            if self.db:
                log.info("using cached database: {db.name}", db=self.db)

            if self.db is None:
                try:
                    self.db = self.db_getter()
                except AttributeError as err:
                    if not self.autoload_db:
                        log.info("no value found: {err}", err=err)
                        return
                    else:
                        log.info("no database provided: {err}", err=err)
                else:
                    log.info("found database: {db.name}", db=self.db)

            # Second: Parse the tags before loading the database.  Loading the 
            # database is expensive, and if the tags won't be in the database 
            # anyways, there's no reason to waste the time.
            
            try:
                tag = self.tag_getter()
            except AttributeError as err:
                log.info("no value found: {err}", err=err)
                return
            else:
                log.info("found tag: {tag!r}", tag=tag)

            try:
                tag = parse_tag(tag)
            except ParseError as err:
                log.info("no value found: not a valid FreezerBox tag")
                return

            # Third: Load the database, if necessary.

            if self.db is None:
                from .model import load_db
                self.db = load_db()
                log.info("loaded database: {db.name}", db=self.db)

            # Fourth: Lookup the key as an attribute of the selected reagents.

            try:
                reagent = self.db[tag]
            except QueryError:
                log.info("no value found: tag not in database")
                return
            else:
                log.info("found reagent: {reagent!r}", reagent=reagent)

            if self.transform_func:
                try:
                    reagent = self.transform_func(reagent)
                except (QueryError, AttributeError) as err:
                    log.info("no value found: {err}", err=err)
                    return
                else:
                    log.info("called: {transform!r}\nreturned: {reagent}", transform=self.transform_func, reagent=reagent)

            yield from getattr_or_call(reagent, key, log)

    def __init__(self, obj, *,
            autoload_db=None,
            db_getter=None,
            tag_getter=None,
            transform=None,
    ):
        super().__init__(obj)

        self.db_getter = db_getter or unbind_method(self.db_getter) 
        self.tag_getter = tag_getter or unbind_method(self.tag_getter) 
        self.transform = transform or unbind_method(self.transform) 

        self.autoload_db = (
                autoload_db if autoload_db is not None else self.autoload_db)

    def load(self):
        yield self.Layer(
                db_getter=self.get_db,
                tag_getter=self.get_tag,
                transform_func=self.transform,
                autoload_db=self.autoload_db,
        )

    def get_db(self):
        return self.db_getter(self.obj)

    def get_tag(self):
        return self.tag_getter(self.obj)


class DeprecatedReagentConfig:
    autoload = True
    autoload_db = True
    db_getter = lambda obj: obj.db
    tag_getter = lambda obj: obj.tag
    transform = list

    class QueryHelper:

        def __init__(self, config, obj):
            self.config = config
            self.obj = obj
            self.db = None

        def __getitem__(self, key):
            # First: See if the object has a database attribute.  If it does, 
            # it costs nothing to access it, and it will allow us to include 
            # the path to the database in error messages if any subsequent 
            # steps fail.

            if self.db is None:
                try:
                    self.db = self.config.db_getter(self.obj)
                except AttributeError:
                    pass

            # Second: Parse the tags before loading the database.  Loading the 
            # database is expensive, and if the tags won't be in the database 
            # anyways, there's no reason to waste the time.
            
            try:
                tags = self.config.tag_getter(self.obj)
            except AttributeError as err:
                raise KeyError from err

            try:
                tags = [
                        parse_tag(x)
                        for x in always_iterable(tags)
                ]
            except ParseError as err:
                raise KeyError from err

            # Third: Load the database.

            if self.db is None and self.config.autoload_db:
                from .model import load_db
                self.db = load_db()

            if self.db is None:
                raise KeyError("no freezerbox database found")

            # Fourth: Lookup the key as an attribute of the selected reagents.

            try:
                values = [
                        key(self.db[x]) if callable(key) else getattr(self.db[x], key)
                        for x in tags
                ]

            except (QueryError, AttributeError) as err:
                raise KeyError from err

            return self.config.transform(values)

        def get_location(self):
            return self.db.name if self.db is not None else "*no database loaded*"

    def __init__(self, tag_getter=None, db_getter=None, autoload_db=None, transform=None):
        cls = self.__class__

        # Access the getter/transform functions through the class.  If accessed 
        # via the instance, they would become bound and would require a self 
        # argument. 

        self.db_getter = db_getter or cls.db_getter
        self.tag_getter = tag_getter or cls.tag_getter
        self.autoload_db = autoload_db if autoload_db is not None else self.autoload_db
        self.transform = transform or cls.transform

    def load(self, obj):
        helper = self.QueryHelper(self, obj)
        yield appcli.Layer(
                values=helper,
                location=helper.get_location,
        )



@autoprop
class ProductConfig(appcli.Config):
    autoload = False
    products_getter = lambda obj: obj.products

    class Layer(appcli.Layer):

        def __init__(self, products_getter):
            self.products_getter = products_getter

        def iter_values(self, key, log):
            try:
                products = self.products_getter()
            except AttributeError as err:
                log.info("no products found: {err}", err=err)
                return

            try:
                product = one(products)
            except ValueError:
                err = QueryError(
                        lambda e: f"expected 1 product, found {len(products)}",
                        products=products,
                )
                raise err from None
            else:
                log.info("found product: {product!r}", product=product)

            yield from self.iter_product_values(product, key, log)

        def iter_product_values(self, product, key, log):
            yield from getattr_or_call(product, key, log)

    def __init__(self, obj, *, products_getter=None):
        super().__init__(obj)
        self.products_getter = \
                products_getter or unbind_method(self.products_getter) 

    def load(self):
        yield self.Layer(self.get_products)

    def get_products(self):
        return self.products_getter(self.obj)


class MakerConfig(ProductConfig):

    class Layer(ProductConfig.Layer):

        def iter_product_values(self, product, key, log):
            dict_layer = appcli.DictLayer(product.maker_args)
            yield from dict_layer.iter_values(key, log)


class PrecursorConfig(ProductConfig):

    class Layer(ProductConfig.Layer):

        def iter_product_values(self, product, key, log):
            precursor = product.precursor
            yield from getattr_or_call(precursor, key, log)

def getattr_or_call(obj, key, log):

    if callable(key):
        try:
            value = key(obj)
        except (QueryError, AttributeError) as err:
            log.info("no value found: {err}", err=err)
            return
        else:
            log.info("called: {key!r}\nreturned: {value!r}", key=key, value=value)

    else:
        try:
            value = getattr(obj, key)
        except (QueryError, AttributeError) as err:
            log.info("no value found: {err}", err=err)
            return
        else:
            log.info("found {key!r}: {value!r}", key=key, value=value)

    yield value



#!/usr/bin/env python3

import autoprop
import entrypoints
from more_itertools import one, only
from dataclasses import dataclass
from collections import namedtuple
from configurator import Config
from voluptuous import Schema
from inform import plural
from Bio.SeqUtils import molecular_weight, MeltingTemp
from .config import load_config
from .errors import LoadError, QueryError, CheckError, only_raise
from .utils import *

DB_PLUGINS = entrypoints.get_group_named('freezerbox.databases')
MAKER_PLUGINS = entrypoints.get_group_named('freezerbox.make')
INTERMEDIATE_SUBCLASSES = {}

class Database:

    def __init__(self, name='*unnamed database*'):
        self.name = name
        self._reagents = {}

    def __iter__(self):
        yield from self._reagents

    def __len__(self):
        return len(self._reagents)

    def __getitem__(self, tag):
        tag = parse_tag(tag)
        try:
            return self._reagents[tag]
        except KeyError:
            raise QueryError(f"not found in database", culprit=tag) from None

    def __setitem__(self, tag, reagent):
        tag = parse_tag(tag)
        if tag in self._reagents:
            raise LoadError("already in database, cannot be replaced", culprit=tag)
        if tag.type != reagent.tag_prefix:
            err = LoadError(
                    tag=tag,
                    reagent=reagent,
            )
            err.brief = "{reagent} cannot have tag '{tag}'"
            err.blame += "expected {reagent.tag_prefix!r} prefix"
            raise err

        self._reagents[tag] = reagent
        reagent._db = self
        reagent._tag = tag

    def __delitem__(self, tag):
        reagent = self._reagents.pop(parse_tag(tag))
        reagent._db = None
        reagent._tag = None
        autoprop.clear_cache(reagent)

    def __contains__(self, reagent):
        return reagent._tag in self._reagents

    def keys(self):
        return self._reagents.keys()

    def values(self):
        return self._reagents.values()

    def items(self):
        return self._reagents.items()


@dataclass(frozen=True)
class Tag:
    type: str
    id: int

    def __str__(self):
        return f'{self.type}{self.id}'


@autoprop.immutable
class Reagent:
    tag_prefix = None

    def __init__(self, **kwargs):
        self._db = None
        self._tag = None
        self._attrs = kwargs
        self._intermediates = {}

    def __repr__(self):
        attrs = list(self._attrs.items())
        if isinstance(self, IntermediateMixin):
            attrs.insert(0, ('step', self.step))

        attr_strs = [f'{k}={v!r}' for k, v in attrs]
        if self._tag:
            attr_strs.insert(0, repr(str(self._tag)))

        return f'{self.__class__.__qualname__}({", ".join(attr_strs)})'

    def check(self):
        pass

    def make_intermediate(self, step):
        if step in self._intermediates:
            return self._intermediates[step]

        if step > len(self.cleanup_args):
            err = QueryError(
                    culprit=self,
                    step=step,
                    synthesis_args=self.synthesis_args,
                    cleanup_args=self.cleanup_args,
            )
            err.brief = "intermediate {step} doesn't exist"
            err.info += lambda e: '\n'.join([
                "intermediates:", *([
                    f'{i}: {x}' 
                    for i, x in enumerate([e.synthesis_args] + e.cleanup_args)
                ]),
            ])
            raise err

        parent = self.parent
        bases = IntermediateMixin, parent.__class__

        try:
            cls = INTERMEDIATE_SUBCLASSES[bases]
        except KeyError:
            name = self.__class__.__name__ + 'Intermediate'
            cls = INTERMEDIATE_SUBCLASSES[bases] = type(name, bases, {})

        intermediate = cls()

        # Note that this is a shallow copy, so changes to any mutable 
        # attributes will be reflected in the intermediate.  That said, 
        # reagents are supposed to be fully immutable, so this should never 
        # matter in practice.

        intermediate.__dict__ = self.__dict__.copy()

        # Set new attributes after overriding `__dict__`:

        intermediate._step = step
        intermediate._parent = parent

        # Any property of the reagent (e.g. concentration, volume, etc.) could 
        # be affected by cleanup steps that come after this intermediate, so 
        # any cached property values need to be discarded.  Properties that 
        # aren't affected by cleanup steps (e.g. seq) can manually implement 
        # caching if desired.

        autoprop.clear_cache(intermediate)

        self._intermediates[step] = intermediate
        return intermediate

    def get_db(self):
        if not self._db:
            raise QueryError("not attached to a database", culprit=self)
        return self._db

    def get_tag(self):
        if not self._tag:
            raise QueryError("not attached to a database", culprit=self)
        return self._tag

    def get_parent(self):
        return self

    def get_name(self):
        return self._attrs.get('name', '')

    def get_alt_names(self):
        return self._attrs.get('alt_names', [])

    def get_date(self):
        return self._attrs.get('date')

    def get_desc(self):
        return self._attrs.get('desc', '')

    def get_dependencies(self):
        try:
            return frozenset(self.synthesis_maker.dependencies)
        except QueryError:
            return frozenset()

    def get_ready(self):
        ready = self._attrs.get('ready', True)

        if callable(ready):
            ready = ready()

        return ready

    def get_maker_attr(self, attr, default=NO_DEFAULT):
        try:
            makers = [self.synthesis_maker, *self.cleanup_makers]
        except QueryError:
            pass
        else:
            for maker in reversed(makers):
                try:
                    return getattr(maker, attr)
                except AttributeError:
                    continue

        if default is not NO_DEFAULT:
            return default
        else:
            raise QueryError(attr)

    def get_synthesis_attr(self, attr, default=NO_DEFAULT):
        try:
            maker = self.synthesis_maker
        except QueryError:
            pass
        else:
            try:
                return getattr(maker, attr)
            except AttributeError:
                pass

        if default is not NO_DEFAULT:
            return default
        else:
            raise QueryError(attr)

    def get_synthesis_maker(self):
        return self.make_intermediate(0).maker

    def get_synthesis_args(self):
        try:
            synthesis = self._attrs['synthesis']
        except KeyError:
            raise QueryError("no synthesis specified", culprit=self) from None

        # Allow `self._synthesis` to be a callable, so that the synthesis 
        # arguments don't have to be loaded until they are actually needed.  
        # This lets the database load faster, and avoids generating errors 
        # relating to synthesis steps that the user doesn't actively care 
        # about.
        if callable(synthesis):
            return synthesis()

        return synthesis

    def get_cleanup_makers(self):
        return [
                self.make_intermediate(i + 1).maker
                for i in range(len(self.cleanup_args))
        ]

    def get_cleanup_args(self):
        # Require that 
        try:
            self._attrs['synthesis']
        except KeyError:
            raise QueryError("no synthesis specified", culprit=self) from None

        cleanups = self._attrs.get('cleanups', [])

        # Allow `self._cleanups` to be a callable, so that the cleanup 
        # arguments don't have to be loaded until they are actually needed.  
        # See `get_synthesis()` for more info.

        if callable(cleanups):
            return cleanups()

        return cleanups

@autoprop.immutable
class Buffer(Reagent):
    tag_prefix = 'b'


@autoprop.immutable
class Molecule(Reagent):
    default_molecule = None

    def __init__(self, **attrs):
        super().__init__(**attrs)
        self._seq = None

    def check(self):
        self._check_seq()

    def get_seq(self):
        # Explicitly cache this property, rather than relying on the caching 
        # provided by autoprop.  We go out of our way to do this because (i) 
        # the sequence is especially expensive to look up and (ii) we know that 
        # it won't be affected by the cleanup steps.

        if self._seq:
            return self._seq

        seq = self._attrs.get('seq')

        # Allow the retrieval of the sequence to be deferred, e.g. so unused 
        # sequences never have to be read from disc.
        if callable(seq):
            seq = seq()

        # If we have instructions for how to make this molecule, try getting
        # the sequence from that.
        if not seq:
            seq = only(
                    self.get_synthesis_attr('product_seqs', []),
                    too_long=QueryError("protocol has multiple sequences", culprit=self),
            )

        if not seq:
            raise QueryError("no sequence specified", culprit=self)

        self._seq = seq
        return seq

    def get_length(self):
        try:
            return self._attrs.get('length') or len(self.seq)
        except QueryError:
            raise QueryError("no length specified", culprit=self)

    def get_mw(self):
        mw = self._attrs.get('mw')

        if not mw:
            mw = self._calc_mw()

        return mw

    def get_conc(self, unit=None):
        conc = self._attrs.get('conc')

        if conc is None:
            conc = self.get_maker_attr('product_conc', None)

        if conc is None:
            raise QueryError("no concentration specified", culprit=self)

        # Allow parsing to be deferred until the concentration is accessed, so 
        # errors don't prevent the rest of the database from loading.
        if callable(conc):
            conc = conc()

        if not unit or unit == conc.unit:
            return conc
        else:
            try:
                mw = self.mw
            except QueryError:
                mw = None

            return convert_conc_unit(conc, mw, unit)

    def get_conc_nM(self):
        return self.get_conc('nM').value

    def get_conc_uM(self):
        return self.get_conc('uM').value

    def get_conc_ng_uL(self):
        return self.get_conc('ng/uL').value

    def get_conc_mg_mL(self):
        return self.get_conc('mg/mL').value

    def get_volume(self):
        volume = self._attrs.get('volume')

        if volume is None:
            volume = self.get_maker_attr('product_volume', None)

        if volume is None:
            raise QueryError("no volume specified", culprit=self)

        # Allow parsing to be deferred until the volume is accessed, so errors 
        # don't prevent the rest of the database from loading.
        if callable(volume):
            volume = volume()

        return volume

    def get_volume_uL(self):
        return self.volume.convert_unit('µL', VOLUME_CONVERSION_FACTORS).value

    def _check_seq(self):
        from Bio import pairwise2
        from Bio.pairwise2 import format_alignment

        try:
            primary_seq = self.seq
            protocol_seq = one(self.get_synthesis_attr('product_seqs'))
        except (QueryError, ValueError):
            pass
        else:
            if primary_seq != protocol_seq:
                alignments = pairwise2.align.globalxx(primary_seq, protocol_seq)
                err = CheckError(culprit=self, alignments=alignments)
                err.brief = "sequence doesn't match construction"
                err.info = lambda e: format_alignment(e.alignments[0])
                raise err

    def _calc_mw(self):
        raise NotImplementedError


@autoprop.immutable
class Protein(Molecule):
    tag_prefix = 'r'

    def get_isoelectric_point(self):
        from Bio.SeqUtils.ProtParam import ProteinAnalysis
        analysis = ProteinAnalysis(self.seq)
        return analysis.isoelectric_point()

    @only_raise(QueryError)
    def _calc_mw(self):
        from Bio.SeqUtils import molecular_weight
        return molecular_weight(
                seq=self.seq,
                seq_type='protein',
        )


@autoprop.immutable
class NucleicAcid(Molecule):
    # "f" for "fragment", i.e. non-plasmid/non-oligo DNA constructs.
    tag_prefix = 'f'
    default_molecule = 'DNA'
    default_strandedness = None

    def __init__(self, **attrs):
        super().__init__(**attrs)
        self._molecule = None
        self._strandedness = None

    def get_molecule(self):
        self._cache_stranded_molecule()
        assert self._molecule
        return self._molecule

    @property
    def is_double_stranded(self):
        self._cache_stranded_molecule()
        assert self._strandedness
        return self._strandedness == 2

    @property
    def is_single_stranded(self):
        return not self.is_double_stranded

    @property
    def is_circular(self):
        circular = self._attrs.get('circular')

        if circular is None:
            circular = self.get_synthesis_attr('is_product_circular', False)

        if circular is None:
            circular = False

        if callable(circular):
            circular = circular()

        return circular

    @property
    def is_linear(self):
        return not self.is_circular

    @property
    def is_phosphorylated_5(self):
        # This attribute is only used when calculating molecular weight.  In 
        # the future, I want to support IDT-style sequence strings, so I could 
        # figure this out by looking for "/5Phos/" or "/3Phos/".  In the 
        # meantime, I'll just assume nothing is phosphorylated.
        return self.get_synthesis_attr('is_product_phosphorylated_5', False)

    @property
    def is_phosphorylated_3(self):
        return self.get_synthesis_attr('is_product_phosphorylated_3', False)

    def _cache_stranded_molecule(self):
        if not self._molecule or not self._strandedness:
            self._parse_stranded_molecule()

    def _parse_stranded_molecule(self):
        molecule = self._attrs.get('molecule')

        if not molecule:
            molecule = self.get_synthesis_attr('product_molecule', None)

        if not molecule:
            molecule = self.default_molecule

        if not molecule:
            raise QueryError("no molecule specified")

        with QueryError.add_info(culprit=self):
            self._molecule, self._strandedness = parse_stranded_molecule(
                    molecule, self.default_strandedness)

    @only_raise(QueryError)
    def _calc_mw(self):
        from Bio.SeqUtils import molecular_weight

        try:
            mw = molecular_weight(
                    seq=self.seq,
                    seq_type=self.molecule,
                    double_stranded=self.is_double_stranded,
                    circular=self.is_circular,
            )
            # For some reason Biopython just assumes 5' phosphorylation, so we 
            # need to correct for that here.
            if not self.is_phosphorylated_5:
                num_strands = 2 if self.is_double_stranded else 1
                num_ends = 0 if self.is_circular else num_strands
                hpo3 = 1.008 + 30.974 + 3*15.999
                mw -= hpo3 * num_ends

            return mw

        except QueryError:
            pass

        try:
            self._cache_stranded_molecule()
            molecule = self._molecule, self._strandedness
            return mw_from_length(self.length, molecule)

        except QueryError:
            pass

        raise QueryError("need sequence or length to calculate molecular weight")


@autoprop.immutable
class Plasmid(NucleicAcid):
    tag_prefix = 'p'

    def get_molecule(self):
        return 'DNA'

    @property
    def is_double_stranded(self):
        return True

    @property
    def is_circular(self):
        return True

@autoprop.immutable
class Oligo(NucleicAcid):
    tag_prefix = 'o'
    default_strandedness = 1

    def get_melting_temp(self):
        from Bio.SeqUtils import MeltingTemp

        # If the Tm is encoded in the oligo name, use that.
        if m := re.search(r'[-_ ](TM|Tm|tm)=?(\d+)', self.name):
            return float(m.group(2))

        # Otherwise, calculate a Tm using the Wallace rule.  This isn't a 
        # particularly accurate method, but I chose it because it agrees most 
        # closely with NEB's Tm calculator, which is what I've been using for 
        # everything.
        else:
            return MeltingTemp.Tm_Wallace(self.seq)

    def get_tm(self):
        return self.melting_temp



class MakerInterface:
    # Maker classes are not required to actually inherit from this class, but 
    # they are expected to implement this interface.

    # Note that I don't use autoprop on this class, because I don't want 
    # getters to be part of the interface.  Subclasses are free to use 
    # autoprop, though.

    @classmethod
    def make(self, db, products):
        raise NotImplementedError

    @property
    def protocol(self):
        raise AttributeError

    @property
    def protocol_duration(self):
        raise AttributeError

    @property
    def dependencies(self):
        raise AttributeError

    @property
    def products(self):
        # Required by make for "Label the product..." step.
        raise AttributeError

    @property
    def product_tags(self):
        raise AttributeError

    @property
    def product_seqs(self):
        raise AttributeError

    @property
    def product_molecule(self):
        raise AttributeError

    @property
    def product_conc(self):
        """
        Return a `Quantity` with one of the following units:
        - "nM"
        - "uM"
        - "µM"
        - "ng/uL"
        - "ng/µL"
        """
        raise AttributeError

    @property
    def product_volume(self):
        """
        Return a `Quantity` with units of "uL" or "µL".
        """
        raise AttributeError

    @property
    def is_product_circular(self):
        raise AttributeError

    @property
    def is_product_phosphorylated_5(self):
        raise AttributeError

    @property
    def is_product_phosphorylated_3(self):
        raise AttributeError

    @property
    def label_products(self):
        raise AttributeError


@autoprop.immutable
class IntermediateMixin:
    # Mixin class; must appear before Reagent in MRO, e.g.:
    #
    # class IntermediatePlasmid(IntermediateMixin, Plasmid):
    #     pass

    def get_step(self):
        # `self._step` is set by `Reagent.make_intermediate()`.
        return self._step

    def get_parent(self):
        return self._parent

    def get_precursor(self):
        if self.step == 0:
            raise QueryError(f"can't get precursor of first intermediate", culprit=self)

        return self.make_intermediate(self.step - 1)

    def get_maker(self):
        try:
            key = self.maker_args[0]
        except IndexError:
            raise QueryError("no protocol specified", culprit=self)

        factory = load_maker_factory(key)
        makers = list(factory(self.db, [self]))
        return one(makers)

    def get_maker_args(self):
        if self.step == 0:
            return self.synthesis_args
        else:
            return self.cleanup_args[-1]

    def get_cleanup_args(self):
        return super().cleanup_args[:self.step]

@only_raise(LoadError)
def load_db(use=None, config=None):
    if not config:
        # Can't test this line, because it reads the real configuration files 
        # on the tester's machine, and so cannot be made to produce consistent 
        # results.
        config = load_config()  # pragma: no cover

    if not use:
        try:
            use = config['use']
        except KeyError as err:
            raise LoadError("no database specified.") from None

    try:
        config_use = config['database'][use]
    except KeyError as err:
        raise LoadError(f"unknown database {use!r}") from None

    try:
        type_use = config_use['type']
    except KeyError as err:
        raise LoadError(f"no 'type' specified for database {use!r}") from None

    try:
        plugin = DB_PLUGINS[type_use].load()
    except KeyError as err:
        raise LoadError(f"no {err} database plugin found") from None

    if defaults := getattr(plugin, 'default_config'):
        config_use = (Config(defaults) + config_use).data

    if hasattr(plugin, 'schema'):
        schema = Schema(plugin.schema)
        config_use = schema(config_use)

    return plugin.load(config_use)

def load_maker_factory(key):
    try:
        plugin = MAKER_PLUGINS[key].load()
    except KeyError as err:
        raise QueryError(f"no {err.args[0]!r} maker plugins found")

    return plugin.make




#!/usr/bin/env python3

import pytest
import freezerbox
import stepwise

class MockReagent(freezerbox.Reagent):
    tag_prefix = 'x'

class MockMolecule(freezerbox.Molecule):
    tag_prefix = 'x'

    def _calc_mw(self):
        raise freezerbox.QueryError

class MockNucleicAcid(freezerbox.NucleicAcid):
    tag_prefix = 'x'

class MockSoloMaker:

    @classmethod
    def make(cls, db, products):
        yield from (cls(x) for x in products)

    def __init__(self, product):
        args = product.maker_args

        self.products = [product]
        self.dependencies = args.get('deps', [])

        if isinstance(self.dependencies, str):
            self.dependencies = self.dependencies.split(',')

        if 'protocol' in args:
            self.protocol = stepwise.Protocol(steps=args['protocol'])

        # Maker attributes:
        if 'conc' in args:
            self.product_conc = stepwise.Quantity.from_string(args['conc'])

        if 'volume' in args:
            self.product_volume = stepwise.Quantity.from_string(args['volume'])

        # Synthesis attributes:
        if 'seq' in args:
            self.product_seqs = [args['seq']]

        if 'molecule' in args:
            self.product_molecule = args['molecule']

        if 'circular' in args:
            self.is_product_circular = freezerbox.parse_bool(args['circular'])


class MockComboMaker:

    @classmethod
    def make(cls, db, products):
        yield cls(list(products))

    def __init__(self, products):
        from more_itertools import flatten

        steps = list(flatten(
            x.maker_args.get('protocol', [])
            for x in products
        ))
        deps = list(flatten(
            x.maker_args.get('deps', [])
            for x in products
        ))

        self.products = products
        self.protocol = stepwise.Protocol(steps=steps)
        self.dependencies = deps
class MockEntryPoint:

    def __init__(self, plugin):
        self.plugin = plugin

    def load(self):
        return self.plugin

MockMaker = MockSoloMaker

@pytest.fixture
def mock_plugins(monkeypatch):
    from string import ascii_lowercase
    monkeypatch.setattr(freezerbox.model, 'MAKER_PLUGINS', {
        **freezerbox.model.MAKER_PLUGINS,
        'mock': MockEntryPoint(MockSoloMaker),
        'merge': MockEntryPoint(MockComboMaker),
        **{k: MockEntryPoint(MockSoloMaker) for k in ascii_lowercase},
    })




#!/usr/bin/env python3

import functools
from tidyexc import Error, only_raise

class Po4Error(Error):

    @property
    def brief_str(self):
        brief_str = super().brief_str
        culprit = self.data.get('culprit')

        if culprit:
            culprit = getattr(culprit, '_tag', None) or culprit
            brief_str = f"{culprit}: {brief_str}"

        return brief_str


class LoadError(Po4Error):
    # For errors relating to loading the database.
    pass

class QueryError(Po4Error):
    # For errors relating to looking up information from the database.
    pass

class ParseError(QueryError):
    # For errors relating to parsing a value from the database, e.g. a unit.
    pass

class CheckError(Po4Error):
    # For errors found when checking the database for self-consistency.
    pass





#!/usr/bin/env python3

import appcli
import autoprop
import stepwise
import freezerbox
import shlex

from freezerbox import (
        load_maker_factory, group_by_synthesis, group_by_cleanup,
        iter_combo_makers, group_by_identity,
        join_lists, join_sets, unanimous, only_raise, QueryError, cd
)
from stepwise import Quantity
from more_itertools import one, first
from operator import not_
from inform import plural
from os import getcwd
from os.path import expanduser

@autoprop.cache
class Make(appcli.App):
    """\
Display a protocol for making the given reagents.

Usage:
    make <tags>... [-R] [-x <tags>]

Arguments:
    <tags>
        The names of any number of reagents in the FreezerBox database, e.g. 
        p01 or f01.  By default, any other reagents that are needed to make the 
        named reagents and that are marked as "not ready" (e.g. "n", "no", "0" 
        in the "Ready" column) will also be included in the resulting protocol.

Options:
    -R --no-recurse
        Only make the reagents specified on the command line; don't 
        automatically include dependencies that are marked as "not ready" in 
        the database.

    -x --exclude <tags>
        A comma-separated list of tags to exclude from the protocol, i.e. 
        dependencies that would normally be included in the protocol but should 
        be excluded for some idiosyncratic reason.

Protocols are derived from the "Synthesis" and "Cleanups" columns of the 
FreezerBox database.  There is an important distinction between these two 
columns.  The "Synthesis" column is for protocols that actually create new 
reagents.  This most often means creating new sequences.  Examples of such 
protocols include PCR, Gibson or Golden Gate assemblies, restriction digests, 
etc.  In contrast, the "Cleanups" column is for protocols that don't create new 
reagents.  These often relate to things like purification or storage.  Examples 
include minipreps, gel purifications, aliquoting, etc.

Each reagent can only have one synthesis protocol, but can have any number of 
cleanup protocols.  This limit on synthesis protocols ensures that every 
reagent has its own name and can be easily and unambiguously referred to in 
other protocols.

Both columns have the same basic syntax for specifying protocol parameters 
(shown below).  For the "Cleanups" column, multiple protocols can be specified 
by separating several of these specifications with semi-colons (;).

    <protocol> [<value>]... [<key>=<value>]...

The following protocols are currently installed:

<%!
from freezerbox.model import MAKER_PLUGINS
from textwrap import indent
%>\
${indent('\\n'.join(MAKER_PLUGINS), '    ')}

Except for `sw` and `order`, each of these should correspond to a stepwise 
protocol of the same (or similar) name.  Information about the parameters 
expected by each protocol can obtained by running that protocol with the `-h` 
flag, e.g. `sw pcr -h`.  Look for a section in the resulting help text labeled 
"Database".  If you are unsure of a specific protocol's name, it may be helpful 
to get a list of every installed protocol (not all of which can be used in the 
FreezerBox database) by running `sw list`.

`sw` and `order` are built-in protocols available only in FreezerBox.  They do 
not correspond to any stepwise commands, but are documented below:

`sw`
    Create reagents using arbitrary stepwise commands.

        sw <command>... [deps=<tags>] [cwd=<path>] [expt=<id>] [project=<path>] 
            [seq=<seq>] [conc=<conc>] [volume=<vol>] [molecule=<type>]

    <command>
        The arguments to pass to stepwise.  Note that this may need to be 
        quoted if characters such as equals (=), semi-colon (;), backslash (\\), 
        or quotes themselves ('") appear in the command.

    deps=<tags>
        The tags (e.g. p1, f1) of any reagents that must be synthesized before 
        the reagent in question.  If not specified, you may get protocols with 
        steps out of order.

    cwd=<path>
        The directory that should be moved to before executing the protocol 
        command.  If not specified, the command will be executed in the current 
        working directory.  This is useful for commands that are not installed 
        globally.

    expt=<id>
        The id number of an Ex Memo experiment.  If specified, the protocol 
        command will be executed from the directory corresponding to that 
        experiment.  This is basically a more succinct way to specify `cwd`, in 
        the event that you use Ex Memo.

    project=<path>
        The path to the root directory of an Ex Memo project.  This is used in 
        conjunction with the `expt` option described above.  If not specified, 
        the project encompassing the current working directory will be used.

    seq=<seq>
        The sequence of the reagent (if applicable).  This can also be 
        specified in the "Sequence" column.

    conc=<conc>
        The concentration of the reagent (if applicable), including a unit.  
        This can also be specified in the "Sequence" column.

    volume=<vol>
        The concentration of the reagent (if applicable), including a unit.  
        This can also be specified in the "Sequence" column.

    molecule=<molecule>
        What kind of nucleic acid the reagent is (if applicable), e.g. DNA, 
        RNA, ssDNA, dsRNA, etc.

    Unlike protocols with devoted plugins, protocols specified in this way 
    can't be smartly merged into succinct master mixes, because FreezerBox 
    doesn't really understand anything about them.  However, it's still useful 
    to be able to specify arbitrary commands for one-off reagents.  If you find 
    yourself using this protocol a lot, you might want to think about writing a 
    protocol plugin.  

    Note that you cannot pipe commands with this protocol.  If you need one or 
    more pipes, put your commands in a shell script and reference that script 
    from the database.

`order`
    Indicate that a reagent was ordered from a vendor.

        order vendor=<name> [seq=<seq>] [conc=<conc>] [volume=<vol>] 
            [molecule=<type>]

    vendor=<name>
        The name of the company the reagent was ordered from.

    seq=<seq>
        The sequence of the reagent (if applicable).  This can also be 
        specified in the "Sequence" column.

    conc=<conc>
        The concentration of the reagent (if applicable), including a unit.  
        This can also be specified in the "Sequence" column.

    volume=<vol>
        The concentration of the reagent (if applicable), including a unit.  
        This can also be specified in the "Sequence" column.

    molecule=<molecule>
        What kind of nucleic acid the reagent is (if applicable), e.g. DNA, 
        RNA, ssDNA, dsRNA, etc.
"""

    __config__ = [
            appcli.DocoptConfig,
    ]

    tags = appcli.param('<tags>')
    recurse_deps = appcli.param('--no-recurse', cast=not_, default=True)
    exclude_deps = appcli.param('--exclude', cast=lambda x: x.split(','), default=frozenset())

    def __init__(self, db, tags=None):
        self.db = db
        self.tags = tags or []

    def get_protocol(self):
        protocol = stepwise.Protocol()
        targets = iter_targets(
                self.db, self.tags,
                recurse_deps=self.recurse_deps,
                exclude_deps=self.exclude_deps,
        )

        for key, group in group_by_synthesis(targets):
            for maker in iter_makers(self.db, key, group):
                protocol += maker.protocol
                if getattr(maker, 'label_products', True):
                    protocol += label_products(maker.products)

            parents = [x.parent for x in group]
            for key, subgroup in group_by_cleanup(parents):
                for maker in iter_makers(self.db, key, subgroup):
                    protocol += maker.protocol

        return protocol

@autoprop
class StepwiseMaker:

    def __init__(self):
        self.dependencies = set()
        self._product_seqs = []
        self._product_concs = []
        self._product_volumes = []
        self._product_molecules = []

    @classmethod
    def make(cls, db, products):
        yield from iter_combo_makers(
                cls,
                map(cls.from_product, products),
                group_by={
                    '_protocol_str': group_by_identity,
                },
                merge_by={
                    'protocol': first,
                    'dependencies': join_sets,
                    '_product_seqs': join_lists,
                    '_product_concs': join_lists,
                    '_product_volumes': join_lists,
                    '_product_molecules': join_lists,
                }
        )

    @classmethod
    def from_product(cls, product):
        maker = cls()
        args = product.maker_args

        maker.products = [product]

        if 'deps' in args:
            maker.dependencies = {x.strip() for x in args['deps'].split(',')}
        if 'seq' in args:
            maker._product_seqs = [args['seq']]
        if 'conc' in args:
            maker._product_concs = [Quantity.from_string(args['conc'])]
        if 'volume' in args:
            maker._product_volumes = [Quantity.from_string(args['volume'])]
        if 'molecule' in args:
            maker._product_molecules = [args['molecule']]

        if 'cwd' in args:
            cwd = expanduser(args['cwd'])
        elif 'expt' in args:
            import exmemo
            root = expanduser(args.get('project', getcwd()))
            work = exmemo.Workspace.from_path(root)
            expt = work.pick_experiment(args['expt'])
            cwd = expt.root_dir.resolve()
        else:
            cwd = getcwd()

        load_cmd = shlex.join(args.by_index[1:])
        if not load_cmd:
            raise QueryError("no stepwise command specified", culprit=product)

        with cd(cwd):
            maker.protocol = stepwise.load(load_cmd).protocol

        maker._protocol_str = maker.protocol.format_text()

        return maker

    def get_product_seqs(self):
        return self._product_seqs

    @only_raise(QueryError, AttributeError)
    def get_product_conc(self):
        return self._unanimous_or_undef(self._product_concs, 'product_conc')

    @only_raise(QueryError, AttributeError)
    def get_product_volume(self):
        return self._unanimous_or_undef(self._product_volumes, 'product_volume')

    @only_raise(QueryError, AttributeError)
    def get_product_molecule(self):
        return self._unanimous_or_undef(self._product_molecules, 'product_molecule')

    def _unanimous_or_undef(self, values, attr):
        return unanimous(
                values,
                err_empty=AttributeError(f"{self!r} object has no attribute {attr!r}"),
                err_multiple=lambda x1, x2: QueryError(f"{self!r} object has multiple values for {attr!r}: {x1!r}, {x2!r}")
        )


@autoprop
class OrderMaker:

    @classmethod
    def make(cls, db, products):
        yield from map(cls.from_product, products)

    @classmethod
    def from_product(cls, product):
        maker = cls()
        args = product.maker_args

        maker.products = [product]
        maker.dependencies = set()
        maker.vendor = args['vendor']
        maker.label_products = False

        if 'seq' in args:
            maker.product_seqs = [args['seq']]
        if 'conc' in args:
            maker.product_conc = Quantity.from_string(args['conc'])
        if 'volume' in args:
            maker.product_volume = Quantity.from_string(args['volume'])
        if 'molecule' in args:
            maker.product_molecule = args['molecule']

        return maker

    def get_protocol(self):
        p = stepwise.Protocol()
        p += f"Order {one(self.products).tag} from {self.vendor}."
        return p

def iter_makers(db, key, targets):
    factory = load_maker_factory(key)
    yield from factory(db, targets)

def iter_targets(db, tags, recurse_deps=True, exclude_deps=frozenset()):
    for tag in tags:
        if tag in exclude_deps:
            continue

        target = db[tag]
        yield target

        if recurse_deps:
            try:
                dep_tags = target.dependencies
            except QueryError:
                continue
            else:
                yield from iter_targets(
                        db, filter(lambda x: not db[x].ready, dep_tags),
                        recurse_deps=recurse_deps,
                        exclude_deps=exclude_deps,
                )

def label_products(products):
    tags = ', '.join(str(x.tag) for x in products)
    return f"Label the {plural(products):product/s}: {tags}"


if __name__ == '__main__':
    app = Make.from_params()
    app.db = freezerbox.load_db()
    app.load()
    app.protocol.print()






#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['freezerbox',
 'freezerbox.cli',
 'freezerbox.group_by',
 'freezerbox.loaders',
 'freezerbox.stepwise']

package_data = \
{'': ['*']}

install_requires = \
['appcli',
 'appdirs',
 'autoprop',
 'autosnapgene',
 'biopython',
 'configurator',
 'docopt',
 'entrypoints',
 'exmemo',
 'inform~=1.21',
 'more_itertools~=8.0',
 'networkx',
 'numpy',
 'pandas',
 'parsy',
 'sklearn',
 'stepwise~=0.23',
 'tidyexc',
 'tqdm',
 'voluptuous']

extras_require = \
{'doc': ['sphinx', 'sphinx_rtd_theme', 'autoclasstoc'],
 'test': ['pytest',
          'pytest-cov',
          'coveralls',
          'parametrize_from_file',
          're_assert']}

entry_points = \
{'console_scripts': ['freezerbox = freezerbox.cli.main:main'],
 'freezerbox.databases': ['excel = freezerbox.loaders.excel'],
 'freezerbox.make': ['dilute = freezerbox.stepwise.dilute:Dilute',
                     'order = freezerbox.stepwise.make:OrderMaker',
                     'sw = freezerbox.stepwise.make:StepwiseMaker'],
 'stepwise.protocols': ['protocols = freezerbox.stepwise:Plugin']}

setup(name='freezerbox',
      version='0.18.0',
      description='Database of DNA sequences.',
      author='Kale Kundert',
      author_email='kale@thekunderts.net',
      url='https://github.com/kalekundert/freezerbox',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='~=3.8',
     )



#!/usr/bin/env python3

from freezerbox import load_db
from schema_helpers import *

@parametrize_from_file(
        schema=Schema({
            'config': {str: eval},
            'error': error,
        }),
)
def test_load_db_err(config, error):
    with error:
        load_db(config=config)
